
open Puzzle0

module TileSet = Set.Make(
  struct 
    type t = tile
    let compare t1 t2 = Pervasives.compare t1.id t2.id
  end)

type t = { 
  board : (tile * int) Pmatrix.t; 
    (* matrix n*n with a tile and an orientation *)
  remaining_corners : TileSet.t;
  remaining_edges : TileSet.t;
  remaining_others : TileSet.t;
  todo  : int; (* number of empty cells *)
  score : int; (* score *)
}

let empty () = 
  let rec loop n (c,e,o) =
    if n<0 then (c,e,o) else
      let p = tiles.(n) in
      match p.kind with
	| Corner _ -> loop (n-1) (TileSet.add p c,e,o)
	| Edge _ -> loop (n-1) (c,TileSet.add p e,o)
	| Other -> loop (n-1) (c,e,TileSet.add p o)
  in
  let (c,e,o) =
    loop (n2-1) (TileSet.empty,TileSet.empty,TileSet.empty)
  in
  assert (TileSet.cardinal c = 4);
  assert (TileSet.cardinal e = 4*(n-2));
  assert (TileSet.cardinal o = (n-2)*(n-2));
{
  board = Pmatrix.init n n (fun _ _ -> (dummy_tile,0));
  remaining_corners = c;
  remaining_edges = e;
  remaining_others = o;
  todo = n2;
  score = 0;
}


let get b i j = Pmatrix.get b.board i j

(* score of tile p (rotation r) if placed at (i,j) *)
let score_delta b i j p r =
  assert (is_real_tile p);
  (if i > 0 then
     let (p0,r0) = get b (i-1) j (* tile below *) in
     if p0.id >= 0 && up_pat p0 r0 = down_pat p r then 1 else 0
   else 0) 
  + 
    (if i < n-1 then
       let (p0,r0) = get b (i+1) j (* tile above *) in
       if p0.id >= 0 && down_pat p0 r0 = up_pat p r then 1 else 0
     else 0) 
  +
    (if j > 0 then
       let (p0,r0) = get b i (j-1) (* tile on the left *) in
       if p0.id >= 0 && right_pat p0 r0 = left_pat p r then 1 else 0
     else 0) 
  +
    (if j < n-1 then
       let (p0,r0) = get b i (j+1) (* tile on the right *) in
       if p0.id >= 0 && left_pat p0 r0 = right_pat p r then 1 else 0
     else 0) 

(* remove tile at (i,j) from b.
   b is modified in-place.
   assumes (i,j) not empty slot *)  
let remove b i j =
  let (p,r) = get b i j in
  let sd = score_delta b i j p r in
  let (c,e,o) = 
    (b.remaining_corners,b.remaining_edges,b.remaining_others) 
  in
  let (c,e,o) =
    match p.kind with
      | Corner _ -> (TileSet.add p c,e,o)
      | Edge _ -> (c,TileSet.add p e,o)
      | Other -> (c,e,TileSet.add p o)
  in
  {
    board = Pmatrix.set b.board i j (dummy_tile,0);
    todo = b.todo + 1;
    score = b.score - sd;
    remaining_corners = c;
    remaining_edges = e;
    remaining_others = o;
  }
  

(* put tile p with rotation r at (i,j) in b.
   b is modified in-place.
   assumes (i,j) empty slot *)  
let put b i j p r =
  let sd = score_delta b i j p r in
  let (c,e,o) = 
    (b.remaining_corners,b.remaining_edges,b.remaining_others) 
  in
  let (c,e,o) =
    match p.kind with
      | Corner _ -> (TileSet.remove p c,e,o)
      | Edge _ -> (c,TileSet.remove p e,o)
      | Other -> (c,e,TileSet.remove p o)
  in
  { 
    board = Pmatrix.set b.board i j (p,r);
    todo = b.todo - 1;
    score = b.score + sd;
    remaining_corners = c;
    remaining_edges = e;
    remaining_others = o;
  }

let initial = 
  let b = empty () in
  let p = tiles.(138) in
  assert (p.id = 139);
  if is_eternity_ii then put b 7 7 tiles.(138) 2 else b 

let create sol =
  let b = ref (empty ()) in
  for i = 0 to n-1 do 
    for j = 0 to n-1 do
      let (p,r) = sol.(i).(j) in
      if p >= 0 then b := put !b i j (tiles.(p-1)) r   
    done 
  done;
  !b

let load f =
  let s = Array.create_matrix n n (-1,0) in
  Solution.load n f s;
  create s 

let save b file =
  let sol = Array.create_matrix n n (-1,0) in
  for i = 0 to n-1 do 
    for j = 0 to n-1 do
      let (p,r) = get b i j in
      sol.(i).(j) <- p.id,r
    done 
  done;
  Solution.save n sol file
