
open Format
open Graphics
open Puzzle

let graphics = ref false
let greedy = ref false
let load = ref None
let sol_dir = ref "best"
let spec = 
  ["-p", Arg.String (fun _ -> ()), "<file>  set puzzle file";
   "-load", Arg.String (fun f -> load := Some f), "<file>  load state";
   "-graphics", Arg.Set graphics, "with graphical display";
   "-greedy", Arg.Set greedy, "search for incomplete solution";
   "-dir", Arg.Set_string sol_dir, "<dir>  directory for solutions";]
let () = Arg.parse spec (fun _ -> ()) 
  "brute.opt [-p puzzle] [-load state] [-graphics] [-greedy]"
let graphics = !graphics
let greedy = !greedy
let load = !load
let sol_dir = !sol_dir

exception Incorrect
exception Incomplete

let raise_incorrect () = raise Incorrect
let raise_incomplete () = if greedy then () else raise Incomplete

let step = 50
let half = step / 2
let scale = half-1

let () =
  if graphics then Draw.open_board step

let draw_tile i j t =
  let x = step + j * step + half in
  let y = step + i * step + half in
  Draw.piece x y scale t

let clear_tile i j =
  let x = step + j * step in
  let y = step + i * step in
  set_color white;
  fill_rect (x+1) (y+1) (step-2) (step-2)

let () = 
  if graphics then begin
    for i = 0 to n-1 do
      for j = 0 to n-1 do
	draw_tile i j pieces.(n*i+j)
      done
    done;
    ignore (read_key ());
    for i = 0 to n-1 do
      for j = 0 to n-1 do
	clear_tile i j
      done
    done
  end

(* TODO: improve using bits of a single int *)
module Todo = 
  Set.Make(struct type t = int * int
		  let compare = Pervasives.compare end)

module Pset = struct
  include 
    Set.Make
    (struct 
       type t = piece
       let compare p1 p2 = Pervasives.compare (p1.id,p1.rot) (p2.id,p2.rot)
     end)

  let of_array a =
    Array.fold_left (fun s p -> add p s) empty a
end

let init_poss =
  Parray.init n 
    (fun i -> 
       Parray.init n 
	 (fun j -> 
	    if i = 0 && j = 0 then
	      2, Pset.of_array corners_rot1
	    else if i = 0 && j = n-1 then
	      2, Pset.of_array corners_rot2
	    else if i = n-1 && j = 0 then
	      2, Pset.of_array corners_rot0
	    else if i = n-1 && j = n-1 then
	      2, Pset.of_array corners_rot3
	    else if i = 0 then
	      1, Pset.of_array edges_rot2
	    else if i = n-1 then
	      1, Pset.of_array edges_rot0
	    else if j = 0 then
	      1, Pset.of_array edges_rot1
	    else if j = n-1 then
	      1, Pset.of_array edges_rot3
	    else
	      0, Pset.union 
		  (Pset.union 
		     (Pset.of_array others_rot0)
		     (Pset.of_array others_rot1))
		  (Pset.union 
		     (Pset.of_array others_rot2)
		     (Pset.of_array others_rot3))
	 ))

let score = ref 0
let best = ref 0

let sol = Array.create_matrix n n dummy_piece

let await = Array.create 23 0
let avail = Array.create 23 0

let () = 
  let add p = avail.(p) <- avail.(p) + 1 in
  Array.iter (fun p -> add p.right; add p.up; add p.left; add p.down) pieces

let add_free p =
  await.(p) <- await.(p) + 1;
  avail.(p) <- avail.(p) - 1;
  if avail.(p) < await.(p) then (printf "!@?"; raise_incomplete ())

let add_bound p =
  await.(p) <- await.(p) - 1;
  avail.(p) <- avail.(p) - 1;
  if avail.(p) < await.(p) then (printf "!@?"; raise_incomplete ())

let remove_free p =
  await.(p) <- await.(p) - 1;
  avail.(p) <- avail.(p) + 1

let remove_bound p =
  await.(p) <- await.(p) + 1;
  avail.(p) <- avail.(p) + 1

module MaxConstr = struct
    (* choose a piece with maximal constraints *)
  let choose _ poss =
    let best_c = ref (-1) in
    let best = ref (-1,0,Pset.empty) in
    let test i j (c,ps) =
      if sol.(i).(j) == dummy_piece && c > !best_c then begin
	best_c := c;
	best := (i,j,ps)
      end
    in
    Parray.iteri (fun i -> Parray.iteri (fun j ps -> test i j ps)) poss;
    match !best with
      | (-1,_,_) -> assert false
      | s -> s
end

module MinCard = struct
    (* choose a piece with minimal possibilities cardinality *)
  let choose _ poss =
    let best_c = ref max_int in
    let best = ref (-1,0,Pset.empty) in
    let test i j (_,ps) =
      let c = Pset.cardinal ps in
      if sol.(i).(j) == dummy_piece && c < !best_c then begin
	best_c := c;
	best := (i,j,ps)
      end
    in
    Parray.iteri (fun i -> Parray.iteri (fun j ps -> test i j ps)) poss;
    match !best with
      | (-1,_,_) -> assert false
      | s -> s
end

module Spiral = struct
    
  let i = ref 7
  let j = ref 7
    
  let pos = Array.create 255 (0,0)
    
  let store =
    let k = ref (-1) in
    fun () -> 
      incr k;
      pos.(!k) <- !i,!j
	
  let up () = incr i
  let down () = decr i
  let right () = incr j
  let left () = decr j
  let repeat n f = for i = 1 to n do f (); store () done
    
  let () =
    for k = 1 to 15 do
      if k mod 2 = 1 then begin
	repeat k up; repeat k right;
      end else begin 
	repeat k down; repeat k left
      end
    done;
    repeat 15 down

  let choose todo poss =
    let i,j = pos.(255-todo) in
    i, j, snd (Parray.get (Parray.get poss i) j)

  let choose1 todo poss = 
    let rec loop k =
      let i,j = pos.(k) in 
      if sol.(i).(j) == dummy_piece then i,j else loop (k+1)
    in
    let i,j = loop 0 in i, j, snd (Parray.get (Parray.get poss i) j)
      
  let choose2 todo poss =
    let rec loop k =
      let i,j = pos.(k) in 
      if sol.(i).(j) == dummy_piece then i,j else loop (k-1)
    in
    let i,j = loop 254 in i, j, snd (Parray.get (Parray.get poss i) j)

  let choose todo poss =
    let i,j =     
      if todo mod 2 = 1 then
	pos.((255-todo) / 2) 
      else
	pos.(254 - (254-todo)/2)
    in
    i, j, snd (Parray.get (Parray.get poss i) j)
      
  let choose todo poss =
    if todo mod 2 = 0 then choose1 todo poss else choose2 todo poss 

end

let dump score =
  let file = Filename.concat sol_dir (string_of_int score ^ ".sol") in
  if is_eternity_ii && not (Sys.file_exists file) then begin
    let s = Array.create_matrix n n (-1,0) in
    for i = 0 to n-1 do 
      for j = 0 to n-1 do
	let p = sol.(i).(j) in
	s.(i).(j) <- (p.id,p.rot)
      done 
    done;    
    Solution.save n s file;
    printf "saved in %s.@." file
  end

let solution () =
  if not greedy then begin
    printf "solution!@.";
    dump 500
  end;
  ignore (read_key ());
  exit 0

let rec fold_n f x =
  let rec loop acc i = if i = n then acc else loop (f acc i) (i+1) in
  loop x 0

(* applies filter [f] on [poss[i,j]] *)
let filter i j f poss =
  let rowi = Parray.get poss i in
  let (c,ps) = Parray.get rowi j in
  let ps = Pset.filter f ps in
  if Pset.is_empty ps then raise_incomplete ();
  Parray.set poss i (Parray.set rowi j (c+1,ps))

(* removes piece [p] from all cells except at (i,j) *)
let filter_p i0 j0 p poss0 = 
  fold_n 
    (fun poss i ->
       let rowi0 = Parray.get poss0 i in
       let rowi = 
	 fold_n 
	   (fun rowi j -> 
	      if is_real_piece sol.(i).(j) then
		rowi
	      else begin
		let (c,ps) = Parray.get rowi0 j in
		let ps = Pset.filter (fun q -> q.id != p.id) ps in
		if (i != i0 || j != j0) && Pset.is_empty ps then 
		  raise_incomplete ();
		Parray.set rowi j (c, ps)
	      end)
	   rowi0
       in
       Parray.set poss i rowi)
    poss0

let place i j p poss =
  let old_score = !score in
  let awr = await.(p.right) in
  let awu = await.(p.up) in
  let awl = await.(p.left) in
  let awd = await.(p.down) in
  let avr = avail.(p.right) in
  let avu = avail.(p.up) in
  let avl = avail.(p.left) in
  let avd = avail.(p.down) in
  if not (is_real_piece p) then begin
    sol.(i).(j) <- p;
    poss     
  end else try
    let poss = filter_p i j p poss in
    let poss = 
      if j > 0 then begin
	let l = sol.(i).(j-1) in
	if not (is_real_piece l) then
	  filter i (j-1) (fun q -> q.right == p.left) poss
	else
	  if l.right == p.left then (incr score; poss) else raise_incorrect ();
      end else
	poss
    in
    let poss = 
      if j < n-1 then begin
	let r = sol.(i).(j+1) in
	if not (is_real_piece r) then
	  filter i (j+1) (fun q -> q.left == p.right) poss
	else
	  if r.left == p.right then (incr score; poss) else raise_incorrect ();
      end else
	poss
    in
    let poss =
      if i > 0 then begin
	let d = sol.(i-1).(j) in
	if not (is_real_piece d) then
	  filter (i-1) j (fun q -> q.up == p.down) poss
	else
	  if d.up == p.down then (incr score; poss) else raise_incorrect ();
      end else
	poss
    in
    let poss = 
      if i < n-1 then begin
	let u = sol.(i+1).(j) in
	if not (is_real_piece u) then
	  filter (i+1) j (fun q -> q.down == p.up) poss
	else
	  if u.down == p.up then (incr score; poss) else raise_incorrect ();
      end else
	poss
    in
    (* await and avail updates *)
    if i > 0 then
      if not (is_real_piece sol.(i-1).(j)) then 
	add_free p.down 
      else
	add_bound p.down;
    if i < n-1 then
       if not (is_real_piece sol.(i+1).(j)) then 
	add_free p.up 
      else
	add_bound p.up;
    if j > 0 then
      if not (is_real_piece sol.(i).(j-1)) then 
	add_free p.left 
      else
	add_bound p.left;
    if j < n-1 then
      if not (is_real_piece sol.(i).(j+1)) then 
	add_free p.right 
      else
	add_bound p.right;
    (* *)
    sol.(i).(j) <- p;
    if graphics then draw_tile i j p;
    if !score > !best then begin
      printf "score = %d@." !score;
      best := !score;
      if !score > 200 then dump !score
    end;
    poss 
  with (Incorrect | Incomplete as exc) ->
    score := old_score;
    await.(p.right) <- awr;
    await.(p.up) <- awu;
    await.(p.left) <- awl;
    await.(p.down) <- awd;
    avail.(p.right) <- avr;
    avail.(p.up) <- avu;
    avail.(p.left) <- avl;
    avail.(p.down) <- avd;
    raise exc

let remove i j =
  let p = sol.(i).(j) in
  sol.(i).(j) <- dummy_piece;
  (* score update *)
  if i > 0 && is_real_piece sol.(i-1).(j) then decr score;
  if i < n-1 && is_real_piece sol.(i+1).(j) then decr score;
  if j > 0 && is_real_piece sol.(i).(j-1) then decr score;
  if j < n-1 && is_real_piece sol.(i).(j+1) then decr score;
  (* await and avail updates *)
  if i > 0 then
    if not (is_real_piece sol.(i-1).(j)) then 
      remove_free p.down 
    else
      remove_bound p.down;
  if i < n-1 then
    if not (is_real_piece sol.(i+1).(j)) then 
      remove_free p.up 
    else
      remove_bound p.up;
  if j > 0 then
    if not (is_real_piece sol.(i).(j-1)) then 
      remove_free p.left 
    else
      remove_bound p.left;
  if j < n-1 then
    if not (is_real_piece sol.(i).(j+1)) then 
      remove_free p.right 
    else
      remove_bound p.right;
  (* graphical display *)
  if graphics then clear_tile i j


let init_poss, init_todo = match load with
  | None ->
      if is_eternity_ii then 
	place 7 7 others_rot2.(139-1-4-56) init_poss, n*n-1
      else 
	init_poss, n*n
  | Some f -> 
      if not (Sys.file_exists f) then begin
	eprintf "cannot open file %s@." f; exit 1
      end;
      let lsol = Puzzle.load f in
      let poss = ref init_poss in
      let todo = ref (n*n) in
      for i = 0 to n-1 do 
	for j = 0 to n-1 do
	  if lsol.(i).(j) != dummy_piece then begin
	    poss := place i j lsol.(i).(j) !poss;
	    decr todo
	  end
	done 
      done;
      !poss, !todo

let rec loop todo poss = 
  if todo = 0 then solution ();
  let i,j,ps = MinCard.choose todo poss in
  Pset.iter
    (fun p -> 
       try 
	 let poss = place i j p poss in 
	 try loop (todo-1) poss with Incorrect | Incomplete -> remove i j
       with Incorrect | Incomplete -> 
	 ())
    ps;
  raise_incomplete ();
  (* only in greedy mode *)
  let poss = place i j dropped_piece poss in 
  loop (todo-1) poss

let () = 
  try loop init_todo init_poss 
  with Incorrect | Incomplete -> printf "No solution!@."

let () = if graphics then ignore (read_key ())
