
open Format
open Graphics
open Puzzle
open Board

let graphics = ref false
let greedy = ref false
let load = ref None
let sol_dir = ref "best2"
let spec = 
  ["-p", Arg.String (fun _ -> ()), "<file>  set puzzle file";
   "-load", Arg.String (fun f -> load := Some f), "<file>  load state";
   "-graphics", Arg.Set graphics, "with graphical display";
   "-greedy", Arg.Set greedy, "search for incomplete solution";
   "-dir", Arg.Set_string sol_dir, "<dir>  directory for solutions";]
let () = Arg.parse spec (fun _ -> ()) 
  "brute.opt [-p puzzle] [-load state] [-graphics] [-greedy]"
let graphics = !graphics
let greedy = !greedy
let load = !load
let sol_dir = !sol_dir

let complete = not greedy

let step = 50
let half = step / 2
let scale = half-1

let () =
  if graphics then Draw.open_board step

let draw_tile i j t =
  let x = step + j * step + half in
  let y = step + i * step + half in
  Draw.piece x y scale t

let clear_tile i j =
  let x = step + j * step in
  let y = step + i * step in
  set_color white;
  fill_rect (x+1) (y+1) (step-2) (step-2)

let iter_grid f =
  for i = 0 to n-1 do for j = 0 to n-1 do f i j done done

let redisplay b =
  iter_grid 
    (fun i j -> 
       let p = piece_at b i j in
       if p != dummy_piece then draw_tile i j p else clear_tile i j)

let () = 
  if graphics then begin
    iter_grid (fun i j -> draw_tile i j pieces.(n*i+j));
    ignore (read_key ());
    iter_grid (fun i j -> clear_tile i j);
    if is_eternity_ii then draw_tile 7 7 (piece_at initial 7 7)
  end

let best = ref 0

module Spiral = struct
    
  let i = ref 7
  let j = ref 7
    
  let pos = Array.create 255 (0,0)
    
  let store =
    let k = ref (-1) in
    fun () -> 
      incr k;
      pos.(!k) <- !i,!j
	
  let up () = incr i
  let down () = decr i
  let right () = incr j
  let left () = decr j
  let repeat n f = for i = 1 to n do f (); store () done
    
  let () =
    for k = 1 to 15 do
      if k mod 2 = 1 then begin
	repeat k up; repeat k right;
      end else begin 
	repeat k down; repeat k left
      end
    done;
    repeat 15 down

  let choose1 b = 
    let rec loop k =
      let i,j = pos.(k) in 
      if piece_at b i j == dummy_piece then i,j else loop (k+1)
    in
    let i,j = loop 0 in i, j, poss b i j
      
  let choose2 b =
    let rec loop k =
      let i,j = pos.(k) in 
      if piece_at b i j == dummy_piece then i,j else loop (k-1)
    in
    let i,j = loop 254 in i, j, poss b i j

  let alternate b =
    if (todo b) mod 2 = 0 then choose1 b else choose2 b 

end

module MinCard = struct
  (* choose a piece with minimal possibilities cardinality *)
  let choose_in f b =
    let best_c = ref max_int in
    let best = ref (-1,0,Pset.empty) in
    let test i j =
      if f i j then
	let ps = poss b i j in
	let c = Pset.cardinal ps in
	if c < !best_c then begin
	  best_c := c;
	  best := (i,j,ps)
	end
    in
    iter_empty test b;
    match !best with
      | (-1,_,_) -> raise Not_found
      | s -> s

  (* choose a piece with minimal possibilities cardinality *)
  let choose b =
    let best_c = ref max_int in
    let best = ref (-1,0,Pset.empty) in
    let test i j =
      let ps = poss b i j in
      let c = Pset.cardinal ps in
      if c < !best_c then begin
	best_c := c;
	best := (i,j,ps)
      end
    in
    iter_empty test b;
    match !best with
      | (-1,_,_) -> assert false
      | s -> s

  let choose2 b =
    if todo b > 230 then Spiral.choose1 b else choose b

  (* choose a piece with minimal possibilities cardinality
     apart from the 3x3 region around (7,7) *)
  let choose_apart77 = 
    choose_in (fun i j -> i < 6 || i > 8 || j < 6 || j > 8)

  let choose3 b =
    if todo b < 20 then choose b else choose_apart77 b

  let avoid = [3,3; 3,11; 11,3; 11,11; 8,8]

  let choose4 b =
    (*try choose_in (fun i j -> j mod 8 = 7) b*)
    (*try choose_in (fun i j -> j <= 8 && i <= 8) b*)
    (*try choose_in (fun i j -> abs (i-7)<=4 && abs(j-7)<=2) b*)
    (*try 
      if (todo b) mod 10 <= 5 then 
	choose_in (fun i j -> i+j <= 12) b
      else
	choose_in (fun i j -> i+j >= 14) b*)
    (*try choose_in (fun i j -> abs (i-7) + abs(j-7) <= 3) b*)
    (*try choose_in (fun i j -> i <= 1 || i >= 14 || j <= 1 || j >= 14) b*)
    (*try choose_in (fun i j -> i <= 0 || i >= 15 || j <= 0 || j >= 15) b*)
    try choose_in (fun i j -> not (List.mem (i,j) avoid)) b
    with Not_found -> choose b

end

let dump b score =
  let file = Filename.concat sol_dir (string_of_int score ^ ".sol") in
  if is_eternity_ii && not (Sys.file_exists file) then begin
    Board.save b file;
    printf "saved in %s.@." file
  end

let solution b =
  if not greedy then begin
    printf "solution!@.";
    dump b 500
  end;
  ignore (read_key ());
  exit 0

let count = ref 1
let limit = ref 100
let limit_step = 100

let place ~complete b i j p =
  let b = Board.place ~complete b i j p in
  if graphics then draw_tile i j p;
  let score = score b in
  if score > !best then begin
    printf "score = %d@." score;
    best := score;
    count := 0;
    if score > 200 then dump b score
  end;
  b

exception Back of int

let rec loop b = 
  if todo b = 0 then solution b;
  incr count; 
  if !count = !limit && todo b < 200 then begin 
    count := 0; limit := 2 * !limit; raise (Back 40) 
  end;
  let i,j,ps = MinCard.choose b in
  Pset.iter
    (fun p -> 
       try 
	 let b = place ~complete b i j p in 
	 try 
	   loop b 
	 with 
	   | Incorrect | Incomplete -> if graphics then clear_tile i j
	   | Back 0 -> printf "!@?"; redisplay b
	   | Back n -> raise (Back (n-1))
       with Incorrect | Incomplete -> 
	 ())
    ps;
  if complete then raise Incomplete;
  (* only in greedy mode *)
  let b = place ~complete:false b i j dropped_piece in 
  loop b

let init_board = match load with
  | None -> 
      Board.initial
  | Some f ->       
      let lsol = Puzzle.load f in
      let b = Board.create ~complete lsol in
      if graphics then redisplay b;
      b

let () = 
  try loop init_board
  with Incorrect | Incomplete -> printf "No solution!@."

let () = if graphics then ignore (read_key ())
