
open Format
open Graphics
open Puzzle
open Board

let graphics = ref false
let greedy = ref false
let load = ref None
let sol_dir = ref "best4"
let spec = 
  ["-p", Arg.String (fun _ -> ()), "<file>  set puzzle file";
   "-load", Arg.String (fun f -> load := Some f), "<file>  load state";
   "-graphics", Arg.Set graphics, "with graphical display";
   "-greedy", Arg.Set greedy, "search for incomplete solution";
   "-dir", Arg.Set_string sol_dir, "<dir>  directory for solutions";]
let () = Arg.parse spec (fun _ -> ()) 
  "brute.opt [-p puzzle] [-load state] [-graphics] [-greedy]"
let graphics = !graphics
let greedy = !greedy
let load = !load
let sol_dir = !sol_dir

let complete = not greedy

let step = 50
let half = step / 2
let scale = half-1

let () =
  if graphics then Draw.open_board step

let draw_tile i j t =
  let x = step + j * step + half in
  let y = step + i * step + half in
  Draw.piece x y scale t

let clear_tile i j =
  let x = step + j * step in
  let y = step + i * step in
  set_color white;
  fill_rect (x+1) (y+1) (step-2) (step-2)

let () = 
  if graphics then begin
    for i = 0 to n-1 do
      for j = 0 to n-1 do
	draw_tile i j pieces.(n*i+j)
      done
    done;
    ignore (read_key ());
    for i = 0 to n-1 do
      for j = 0 to n-1 do
	clear_tile i j
      done
    done;
    if is_eternity_ii then 
      begin
	draw_tile 7 7 (piece_at initial 7 7);
	draw_tile 2 2 (piece_at initial 2 2);
      end
  end

let best = ref 0

module Spiral = struct
    
  let i = ref 7
  let j = ref 7
    
  let pos = Array.create 255 (0,0)
    
  let store =
    let k = ref (-1) in
    fun () -> 
      incr k;
      pos.(!k) <- !i,!j
	
  let up () = incr i
  let down () = decr i
  let right () = incr j
  let left () = decr j
  let repeat n f = for i = 1 to n do f (); store () done
    
  let () =
    for k = 1 to 15 do
      if k mod 2 = 1 then begin
	repeat k up; repeat k right;
      end else begin 
	repeat k down; repeat k left
      end
    done;
    repeat 15 down

  let choose1 b = 
    let rec loop k =
      let i,j = pos.(k) in 
      if piece_at b i j == dummy_piece then i,j else loop (k+1)
    in
    let i,j = loop 0 in i, j, poss b i j
      
  let choose2 b =
    let rec loop k =
      let i,j = pos.(k) in 
      if piece_at b i j == dummy_piece then i,j else loop (k-1)
    in
    let i,j = loop 254 in i, j, poss b i j

  let alternate b =
    if (todo b) mod 2 = 0 then choose1 b else choose2 b 

end

module MinCard = struct
  (* choose a piece with minimal possibilities cardinality *)
  let choose_in f b =
    let best_c = ref max_int in
    let best = ref (-1,0,Pset.empty) in
    let test i j =
      if f i j then
	let ps = poss b i j in
	let c = Pset.cardinal ps in
	if c < !best_c then begin
	  best_c := c;
	  best := (i,j,ps)
	end
    in
    iter_empty test b;
    match !best with
      | (-1,_,_) -> raise Not_found
      | s -> s

  (* choose a piece with minimal distance to corners *)
  let choose_close_to_corners f b =
    let best_c = ref max_int in
    let best = ref (-1,0,Pset.empty) in
    let test i j =
      if f i j then
	let ps = poss b i j in
	let c = Pset.cardinal ps in
	let c = 5*c + 3*corner_distance i j in
	if c < !best_c then begin
	  best_c := c;
	  best := (i,j,ps)
	end
    in
    iter_empty test b;
    match !best with
      | (-1,_,_) -> raise Not_found
      | s -> s

  (* choose a piece with minimal possibilities cardinality *)
  let choose b =
    let best_c = ref max_int in
    let best = ref (-1,0,Pset.empty) in
    let test i j =
      let ps = poss b i j in
      let c = Pset.cardinal ps in
      if c < !best_c then begin
	best_c := c;
	best := (i,j,ps)
      end
    in
    iter_empty test b;
    match !best with
      | (-1,_,_) -> assert false
      | s -> s

  let choose2 b =
    if todo b > 230 then Spiral.choose1 b else choose b

  (* choose a piece with minimal possibilities cardinality
     apart from the 3x3 region around (7,7) *)
  let choose_apart77 = 
    choose_in (fun i j -> i < 6 || i > 8 || j < 6 || j > 8)

  let choose3 b =
    if todo b < 20 then choose b else choose_apart77 b

  let avoid = [3,3; 3,11; 11,3; 11,11; 8,8]

  let choose4 b =
    (*try choose_in (fun i j -> j mod 8 = 7) b*)
    (*try choose_in (fun i j -> j <= 8 && i <= 8) b*)
    (*try choose_in (fun i j -> abs (i-7)<=4 && abs(j-7)<=2) b*)
    (*try 
      if (todo b) mod 10 <= 5 then 
	choose_in (fun i j -> i+j <= 12) b
      else
	choose_in (fun i j -> i+j >= 14) b*)
    (*try choose_in (fun i j -> abs (i-7) + abs(j-7) <= 3) b*)
    (*try choose_in (fun i j -> i <= 1 || i >= 14 || j <= 1 || j >= 14) b*)
    (*try choose_in (fun i j -> i <= 0 || i >= 15 || j <= 0 || j >= 15) b*)
    (*try choose_in (fun i j -> not (List.mem (i,j) avoid)) b*)
    (*try choose_in (fun i j -> abs (i-7)>2 || abs(j-7)>2) b*)
    try choose_in (fun i j -> i=0 || i=15 || j=0 || j=15) b
    with Not_found ->
    try choose_in (fun i j -> i=1 || i=14 || j=1 || j=14) b
    (*try choose_in (fun i j -> not (List.mem (i,j) avoid)) b*)
    (*try choose_in (fun i j -> abs (i-7)>2 || abs(j-7)>2) b*)
    with Not_found -> choose b

  let connected b i j =
    (i > 0 && is_real_piece (piece_at b (i-1) j)) ||
      (i < n-1 && is_real_piece (piece_at b (i+1) j)) ||
      (j > 0 && is_real_piece (piece_at b i (j-1))) ||
      (j < n-1 && is_real_piece (piece_at b i (j+1))) 

  let choose_connected b =
    choose_in (connected b) b

  let choose_connected_close_to_corners b =
    choose_close_to_corners (connected b) b

  let choose_connected_ondiag_if_possible b =
    try 
      choose_in 
	(fun i j -> 
	   0 <= j - i && j - i <= 1 && connected b i j) b
    with
	Not_found -> choose_in (connected b) b

end

let dump b score =
  let file = Filename.concat sol_dir (string_of_int score ^ ".sol") in
  if is_eternity_ii && not (Sys.file_exists file) then begin
    Board.save b file;
    printf "saved in %s.@." file
  end

let nsol = ref 0

let solution b =
  incr nsol;
  if not greedy then begin
    printf "solution %d!@." !nsol;
    dump b 500;
    if graphics then ignore (read_key ())
  end;
  () (* exit 0 *)

let rose = 1
let jaune = 2
let ciel = 3
let orange = 4
let marron = 5
let violet = 6
let marine = 7
let vert = 8

let bg_color = function
  | 0 (* bord *) -> 0
  | 1 (* chateau_jaune *) -> rose
  | 15 (* trefle_jaune *) -> rose

  | 2 (* chateau_bleu *) -> jaune
  | 11 (* carre_vert *) -> jaune
  | 5 (* etoile_bleue *) -> jaune

  | 3 (* chateau_rose *) -> ciel
  | 17 (* cristal_rose *) -> ciel

  | 4 (* etoile_violette *) -> orange

  | 6 (* etoile_jaune *) -> marron
  | 9 (* croix_marron *) -> marron

  | 7 (* croix_violette *) -> violet
  | 18 (* cristal_bleu *) -> violet

  | 8 (* croix_bleue *) -> marine
  | 16 (* trefle_orange *) -> marine
  | 10 (* carre_bleu *) -> marine

  | 14 (* trefle_rose *) -> vert
  | 19 (* cristal_orange *) -> vert

  | 12 (* biscuit *) -> marron
  | 13 (* fleur *) -> marine
  | 20 (* shuriken *) -> orange
  | 21 (* diamand *) -> rose
  | 22 (* ecrou *) -> vert

  | _ -> assert false


let place ~complete b i j p =
  let b = Board.place ~complete b i j p in
  let score = score b in
  if score > !best then begin
    printf "score = %d@." score;
    best := score;
    if score > 300 then dump b score
  end;
  b

let count = ref 0

(* look for two adjacent empty cells with incompatible possibilities *)
let check b =
  iter_empty
    (fun i j ->
       let pij = poss b i j in
       let test pr ps = Pset.for_all (fun p -> Pset.exists (pr p) ps) pij in
       if Pset.cardinal pij < 10 then begin
	 if i > 0 && piece_at b (i-1) j == dummy_piece then begin 
	   if not (test (fun p p' -> p.down == p'.up) (poss b (i-1) j)) then 
	     (printf "!@?"; raise Incomplete)
	 end;
	 if i < n-1 && piece_at b (i+1) j == dummy_piece then begin
	   if not (test (fun p p' -> p.up == p'.down) (poss b (i+1) j)) then
	     (printf "!@?"; raise Incomplete)
	 end;
	 if j > 0 && piece_at b i (j-1) == dummy_piece then begin
	   if not (test (fun p p' -> p.left == p'.right) (poss b i (j-1))) then
	     (printf "!@?"; raise Incomplete)
	 end;
	 if j < n-1 && piece_at b i (j+1) == dummy_piece then begin
	   if not (test (fun p p' -> p.right == p'.left) (poss b i (j+1))) then
	     (printf "!@?"; raise Incomplete)
	 end;
       end)
    b

let () = Random.self_init ()

let isqrt i =
  if i <= 0 then 0
  else if i <= 1 then 1 
  else if i <= 4 then 2 
  else if i <= 9 then 3
  else if i <= 16 then 4
  else if i <= 25 then 5
  else if i <= 36 then 6
  else if i <= 49 then 7
  else if i <= 64 then 8
  else if i <= 81 then 9
  else if i <= 100 then 10
  else if i <= 121 then 11
  else if i <= 144 then 12
  else if i <= 169 then 13
  else 14

let decr_avail avail p =    
  let avail_p = Parray.get avail p - 1 in
  Parray.set avail p avail_p

(* variance of an array

  sum = sum_{i}  t[i]

  moy = sum / n

  var = sum_{i} (t[i]-moy)^2
      = sum_{i} (t[i]^2 - 2 * t[i] * moy + moy ^ 2)
      = sum_{i} t[i]^2 - 2 * moy * sum + n * moy ^ 2 
      = sum_{i} t[i]^2 - 2 * sum^2 / n  + sum ^ 2 / n 
      = sum_{i} t[i]^2 - sum^2 / n   

*)
let variance avail : int =
  let sum = ref 0 in
  let squares = ref 0 in
  Parray.iter
    (fun v ->
       sum := !sum + v;
       squares := !squares + v * v) avail;
  23 * !squares - !sum * !sum

module IntMap = 
  Map.Make(struct
	     type t = int
	     let compare = Pervasives.compare 
	   end)

      
    
let rec loop b = 
  if todo b = 0 then (solution b; raise Incorrect);
  (*incr count; if !count = 50 then begin count := 0; check b end;*)
  let i,j,ps = 
    if todo b = n2
    then MinCard.choose b 
    else MinCard.choose_connected_ondiag_if_possible (*close_to_corners*) b 
  in
(*
(* strategy 0: random *)
  Pset.random_iter
    (fun p -> 
       try 
	 let b = place ~complete b i j p in 
	 if graphics then draw_tile i j p;
	 try loop b 
	 with Incorrect | Incomplete -> if graphics then clear_tile i j
       with Incorrect | Incomplete -> 
	 ())
    ps;
*)

(* 
  (* heuristics 1 : place the tile that leaves maximal possibilities 
   on remaining empty slots  
*)

  let t = Hashtbl.create 97 in
  let () = Pset.iter
    (fun p ->
       try 
	 let b = place ~complete b i j p in 
	 let possibilities = ref 0 in
	 iter_empty
	   (fun i j -> possibilities := !possibilities + 
	      isqrt (Pset.cardinal (poss b i j))) b;
	 let x =
	   try
	     Hashtbl.find t !possibilities
	   with Not_found -> Pset.empty
	 in
	 Hashtbl.replace t !possibilities (Pset.add p x)
       with Incorrect | Incomplete -> ())
    ps 
  in
  let l = Hashtbl.fold (fun i p acc -> (i,p)::acc) t [] in
  let l = List.sort (fun (i1,p1) (i2,p2) -> Pervasives.compare i2 i1) l in
  List.iter
    (fun (_,ps) -> 
       Pset.random_iter 
	 (fun p -> 
	    try
	      let b = place ~complete b i j p in 
	      if graphics then draw_tile i j p;
	      try loop b 
	      with Incorrect | Incomplete -> if graphics then clear_tile i j
	    with Incorrect | Incomplete -> assert false)
	 ps)
    l;
*)
(* heuristics 2: place the tile that leaves the most uniform repartition 
   of remaing patterns *)
  let t = ref IntMap.empty in
  let avail = Board.avail b in
  let () = Pset.iter
    (fun p ->
       let a = decr_avail avail p.up in
       let a = decr_avail a p.left in
       let a = decr_avail a p.down in
       let a = decr_avail a p.right in
       let var = variance a in
       let x =
	 try
	   IntMap.find var !t
	 with Not_found -> Pset.empty
       in
       t := IntMap.add var (Pset.add p x) !t)
    ps 
  in
  IntMap.iter
    (fun _ ps -> 
       Pset.random_iter 
	 (fun p -> 
	    try
	      let b = place ~complete b i j p in 
	      if graphics then draw_tile i j p;
	      try loop b 
	      with Incorrect | Incomplete -> if graphics then clear_tile i j
	    with Incorrect | Incomplete -> ())
	 ps)
    !t;

(* end of heuristics *)
  if complete then raise Incomplete;
  (* only in greedy mode *)
  let b = place ~complete:false b i j dropped_piece in 
  loop b

let iter_grid f =
  for i = 0 to n-1 do for j = 0 to n-1 do f i j done done

let init_board = match load with
  | None -> 
      Board.initial
  | Some f ->       
      let s = Puzzle.load f in
      let b = Board.create ~complete:true s in
      if graphics then begin
	iter_grid 
	  (fun i j -> 
	     let p = piece_at b i j in
	     if p != dummy_piece then draw_tile i j p else clear_tile i j)
      end;
      b

let () = 
  try loop init_board
  with Incorrect | Incomplete -> printf "No solution!@."

let () = if graphics then ignore (read_key ())
