
open Format
open Graphics
open Puzzle
open Board

let graphics = ref false
let resort = ref false
let load = ref None
let sol_dir = ref "best6"
let credit = ref 1
let spec = 
  ["-p", Arg.String (fun _ -> ()), "<file>  set puzzle file";
   "-credit", Arg.Int (fun n -> credit := n), "<n>  credit for leaves";
   "-load", Arg.String (fun f -> load := Some f), "<file>  load state";
   "-graphics", Arg.Set graphics, "with graphical display";
   "-resort", Arg.Set resort, "resorts the tree read from file";
   "-dir", Arg.Set_string sol_dir, "<dir>  directory for solutions";]
let () = Arg.parse spec (fun _ -> ()) 
  "brute5.opt [-p puzzle] [-load state] [-graphics] [-resort]"
let graphics = !graphics
let credit = !credit
let resort = !resort
let load = !load
let sol_dir = !sol_dir


let step = 50
let half = step / 2
let scale = half-1

let () =
  if graphics then Draw.open_board step

let draw_tile i j t =
  let x = step + j * step + half in
  let y = step + i * step + half in
  Draw.piece x y scale t

let clear_tile i j =
  let x = step + j * step in
  let y = step + i * step in
  set_color white;
  fill_rect (x+1) (y+1) (step-2) (step-2)

let border_tile c i j =
  let x = step + j * step in
  let y = step + i * step in
  set_line_width 3;
  set_color c;
  draw_rect (x+2) (y+2) (step-4) (step-4)

let () = 
  if graphics then begin
    for i = 0 to n-1 do
      for j = 0 to n-1 do
	draw_tile i j pieces.(n*i+j)
      done
    done;
    ignore (read_key ());
    for i = 0 to n-1 do
      for j = 0 to n-1 do
	clear_tile i j
      done
    done;
    if is_eternity_ii then 
      begin
	draw_tile 7 7 (piece_at initial 7 7);
	draw_tile 2 2 (piece_at initial 2 2);
      end
  end

let best = ref 0


module MinCard = struct
  (* choose a piece with minimal possibilities cardinality *)
  let choose_in f b =
    let best_c = ref max_int in
    let best = ref (-1,0,Pset.empty) in
    let test i j =
      if f i j then
	let ps = poss b i j in
	let c = Pset.cardinal ps in
	if c < !best_c then begin
	  best_c := c;
	  best := (i,j,ps)
	end
    in
    iter_empty test b;
    match !best with
      | (-1,_,_) -> raise Not_found
      | s -> s

  (* choose a piece with minimal distance to corners *)
  let choose_close_to_corners f b =
    let best_c = ref max_int in
    let best = ref (-1,0,Pset.empty) in
    let test i j =
      if f i j then
	let ps = poss b i j in
	let c = Pset.cardinal ps in
	let c = 5*c + 3*corner_distance i j in
	if c < !best_c then begin
	  best_c := c;
	  best := (i,j,ps)
	end
    in
    iter_empty test b;
    match !best with
      | (-1,_,_) -> raise Not_found
      | s -> s

  (* choose a piece with minimal possibilities cardinality *)
  let choose b =
    let best_c = ref max_int in
    let best = ref (-1,0,Pset.empty) in
    let test i j =
      let ps = poss b i j in
      let c = Pset.cardinal ps in
      if c < !best_c then begin
	best_c := c;
	best := (i,j,ps)
      end
    in
    iter_empty test b;
    match !best with
      | (-1,_,_) -> assert false
      | s -> s

  (* choose a piece with minimal possibilities cardinality
     apart from the 3x3 region around (7,7) *)
  let choose_apart77 = 
    choose_in (fun i j -> i < 6 || i > 8 || j < 6 || j > 8)

  let choose3 b =
    if todo b < 20 then choose b else choose_apart77 b

  let avoid = [3,3; 3,11; 11,3; 11,11; 8,8]

  let choose4 b =
    (*try choose_in (fun i j -> j mod 8 = 7) b*)
    (*try choose_in (fun i j -> j <= 8 && i <= 8) b*)
    (*try choose_in (fun i j -> abs (i-7)<=4 && abs(j-7)<=2) b*)
    (*try 
      if (todo b) mod 10 <= 5 then 
	choose_in (fun i j -> i+j <= 12) b
      else
	choose_in (fun i j -> i+j >= 14) b*)
    (*try choose_in (fun i j -> abs (i-7) + abs(j-7) <= 3) b*)
    (*try choose_in (fun i j -> i <= 1 || i >= 14 || j <= 1 || j >= 14) b*)
    (*try choose_in (fun i j -> i <= 0 || i >= 15 || j <= 0 || j >= 15) b*)
    (*try choose_in (fun i j -> not (List.mem (i,j) avoid)) b*)
    (*try choose_in (fun i j -> abs (i-7)>2 || abs(j-7)>2) b*)
    try choose_in (fun i j -> i=0 || i=15 || j=0 || j=15) b
    with Not_found ->
    try choose_in (fun i j -> i=1 || i=14 || j=1 || j=14) b
    (*try choose_in (fun i j -> not (List.mem (i,j) avoid)) b*)
    (*try choose_in (fun i j -> abs (i-7)>2 || abs(j-7)>2) b*)
    with Not_found -> choose b

  let choose_connected b =
    let connected i j =
      (i > 0 && is_real_piece (piece_at b (i-1) j)) ||
	(i < n-1 && is_real_piece (piece_at b (i+1) j)) ||
	(j > 0 && is_real_piece (piece_at b i (j-1))) ||
	(j < n-1 && is_real_piece (piece_at b i (j+1))) 
    in
    choose_in connected b

  let choose_connected_close_to_corners b =
    let connected i j =
      (i > 0 && is_real_piece (piece_at b (i-1) j)) ||
	(i < n-1 && is_real_piece (piece_at b (i+1) j)) ||
	(j > 0 && is_real_piece (piece_at b i (j-1))) ||
	(j < n-1 && is_real_piece (piece_at b i (j+1))) 
    in
    choose_close_to_corners connected b
end

let dump b score =
  let file = Filename.concat sol_dir (string_of_int score ^ ".sol") in
  if is_eternity_ii && not (Sys.file_exists file) then begin
    Board.save b file;
    printf "saved in %s.@." file
  end

let nsol = ref 0

let solution b =
  incr nsol;
  printf "solution %d!@." !nsol;
  dump b 500;
  () (* exit 0 *)


let place ~complete b i j p =
  let b = Board.place ~complete b i j p in
  let score = score b in
  if score > !best then begin
    printf "score = %d@." score;
    best := score;
    if score > 300 then dump b score
  end;
  b

let count = ref 0


let isqrt i =
  if i <= 0 then 0
  else if i <= 1 then 1 
  else if i <= 4 then 2 
  else if i <= 9 then 3
  else if i <= 16 then 4
  else if i <= 25 then 5
  else if i <= 36 then 6
  else if i <= 49 then 7
  else if i <= 64 then 8
  else if i <= 81 then 9
  else if i <= 100 then 10
  else if i <= 121 then 11
  else if i <= 144 then 12
  else if i <= 169 then 13
  else 14

module IntMap = 
  Map.Make(struct
	     type t = int
	     let compare i1 i2 = Pervasives.compare i2 i1
	   end)

let decr_avail avail p =    
  let avail_p = Parray.get avail p - 1 in
  Parray.set avail p avail_p

(* variance of an array

  sum = sum_{i}  t[i]

  moy = sum / n

  var = sum_{i} (t[i]-moy)^2
      = sum_{i} (t[i]^2 - 2 * t[i] * moy + moy ^ 2)
      = sum_{i} t[i]^2 - 2 * moy * sum + n * moy ^ 2 
      = sum_{i} t[i]^2 - 2 * sum^2 / n  + sum ^ 2 / n 
      = sum_{i} t[i]^2 - sum^2 / n   

*)
let variance avail : int =
  let sum = ref 0 in
  let squares = ref 0 in
  Parray.iter
    (fun v ->
       sum := !sum + v;
       squares := !squares + v * v) avail;
  23 * !squares - !sum * !sum

let rec loop credit b = 
  if todo b = 0 then (solution b; raise Incorrect);
  (*incr count; if !count = 50 then begin count := 0; check b end;*)
  let i,j,ps = 
    if todo b = n2
    then MinCard.choose b 
    else MinCard.choose_connected_close_to_corners b 
  in
(* heuristics 1: maximizes remaining possibilities *)
(*
  let t = ref IntMap.empty in
  let () = Pset.iter
    (fun p ->
       try 
	 let b = place ~complete:true b i j p in 
	 let possibilities = ref 0 in
	 iter_empty
	   (fun i j -> possibilities := !possibilities + 
	      isqrt (Pset.cardinal (poss b i j))) b;
	 let x =
	   try
	     IntMap.find !possibilities !t
	   with Not_found -> Pset.empty
	 in
	 t := IntMap.add !possibilities (Pset.add p x) !t 
       with Incorrect | Incomplete -> ())
    ps 
  in
  let c = ref credit in
  let sc = Board.score b in
  let s = ref sc in
  begin
    try
      IntMap.iter
	(fun _ ps -> 
	   Pset.iter 
	     (fun p -> 
		try 
		  let cr = !c in
		  decr c;
		  if !c < 0 then raise Exit;
		  let b1 = place ~complete:true b i j p in 
		  if graphics then draw_tile i j p;
		  let cr = 
		    let sc1 = Board.score b1 in
		    if sc1 >= 150 && sc / 50 < sc1 / 50 then cr+1 else cr
		  in
		  let sc = loop cr b1 in
		  if sc > !s then s := sc;
		  if graphics then clear_tile i j
		with Incorrect | Incomplete -> ())
	     ps)
	!t;
    with Exit -> ();
  end;
*)
(* heuristics 2: minimizes variance of remaining tiles *)
  let avail = Board.avail b in
  let l = Pset.fold
    (fun p acc ->
       let a = decr_avail avail p.up in
       let a = decr_avail a p.left in
       let a = decr_avail a p.down in
       let a = decr_avail a p.right in
       let var = variance a in
       (var,p)::acc)
    ps []
  in
  let l = List.sort (fun (var1,p1) (var2,p2) -> Pervasives.compare var1 var2) l in
  let c = ref credit in
  let sc = Board.score b in
  let s = ref sc in
  begin
    try
      List.iter
	(fun (_,p) -> 
	   try 
	     let cr = !c in
	     decr c;
	     if !c < 0 then raise Exit;
	     let b1 = place ~complete:true b i j p in 
	     if graphics then draw_tile i j p;
	     let cr = 
	       let sc1 = Board.score b1 in
	       if sc1 >= 150 && sc / 50 < sc1 / 50 then cr+1 else cr
	     in
	     let sc = loop cr b1 in
	     if sc > !s then s := sc;
	     if graphics then clear_tile i j
	   with Incorrect | Incomplete -> ())
	l;
    with Exit -> ();
  end;
(* returns score *)
  !s

let iter_grid f =
  for i = 0 to n-1 do for j = 0 to n-1 do f i j done done

let init_board = Board.initial

type search_tree =
  | Unsolvable
  | Leaf of int * int 
      (* score, credit *)
  | Node of int * int * int * int * (Puzzle.piece * search_tree) list
      (* score, i, j, size, children *)

let score = function
  | Unsolvable -> 0
  | Leaf(s,_) -> s
  | Node(s,_,_,_,_) -> s

let size = function
  | Unsolvable -> max_int
  | Leaf _ -> 0
  | Node(_,_,_,s,_) -> s

(* make a leaf with credit *)
let make_leaf b =
  let sc = Board.score b in
  let c = if sc >= 150 then credit + sc / 50 - 2 else credit in
  let s = loop c b in
  Leaf(s,c)

let rec insert p t s siz l =
  match l with
    | [] -> [(p,t)]
    | (p',t')::r ->
	let s' = score t' in
	let siz' = size t' in
	if (* siz < 3 ||
	      (siz' >= 3 && *) s - s' >= (siz - siz') (* ) *)
	then (p,t)::l
	else (p',t')::(insert p t s siz r)
	
let rec expand_tree b t =
  match t with
    | Unsolvable -> assert false
    | Leaf(s,c) ->
	let i,j,ps = 
	  if todo b = n2
	  then MinCard.choose b 
	  else MinCard.choose_connected_close_to_corners b 
	in
	let l = Pset.fold
	  (fun p acc ->
	     try 
	       let b = place ~complete:true b i j p in 
	       if graphics then (draw_tile i j p; border_tile red i j); 
	       let l = make_leaf b in
	       if graphics then clear_tile i j;
	       (p,l)::acc
	     with Incorrect | Incomplete -> acc)
	  ps []
	in
	if l = [] then Unsolvable else
	  let l = List.sort 
	    (fun (p1,t1) (p2,t2) -> Pervasives.compare (score t2) (score t1)) l 
	  in
	  Node(score (snd (List.hd l)), i, j, 1, l)
    | Node(s,i,j,_,[]) -> assert false
    | Node(s,i,j,siz,(p,t)::l) -> 
	let b = place ~complete:true b i j p in 
	if graphics then (draw_tile i j p; border_tile blue i j);
	let t = expand_tree b t in
	if graphics then clear_tile i j;
	let sc = 
	  List.fold_left
	    (fun acc (_,t) -> max acc (score t))
	    (score t)
	    l
	in
	let siz' = size t in
	Node(sc,i,j,siz+1,insert p t (score t) siz' l)
	
	
let dump_tree file t =
  let ch = open_out file in
  let rec loop fmt t =
    match t with
      | Unsolvable -> fprintf fmt "U"
      | Leaf(s,c) -> fprintf fmt "L(%d,%d)" s c
      | Node(s,i,j,siz,l) ->
	  fprintf fmt "@[N(%d,%d,%d,%d,@,@[[" s i j siz;
	  List.iter
	    (fun (p,t) -> fprintf fmt "(%d,@,%d,@,%a);@ " p.id p.rot loop t)
	    l;
	  fprintf fmt "]@])@]"
  in
  loop (formatter_of_out_channel ch) t;
  close_out ch

let rec build_tree (* b *) a =
  match a with
    | Ast.U -> Unsolvable
    | Ast.L(score,credit,_) -> Leaf(score,credit)
    | Ast.N(scor,i,j,siz,l) ->
	let (sc,si,l) =
	  List.fold_right
	    (fun (p,r,a) (sc,siz,l) ->	       
	       let p = Puzzle.get_piece p r in
	       (* let b = place ~complete:true b i j p in *)
	       let t = build_tree (* b *) a in
	       let sco = score t in
	       let si = size t in
	       (max sc sco,siz+si,
		if resort then insert p t sco si l else (p,t)::l))
	    l (0,1,[])
	in
	if sc <> scor then 
	    failwith "bad score" 
	else
	  if si <> siz then 
	    begin
	      printf "bad size: expected %d, got %d@." si siz;
	      failwith "bad size"
	    end
	  else
	    Node(scor,i,j,siz,l)

let report_iteration b =	       
  let r = 
    "iteration=" ^ (string_of_int (size b)) ^ 
    " best=" ^ (string_of_int (score b))
  in
  if graphics then 
    set_window_title 
      ("Eternity II, brute6: " ^ r)
  else
    printf "%s@." r

let read_tree file =
  let ch = open_in file in
  let b = Lexing.from_channel ch in
  let a = Parse_tree.tree Lex_tree.token b in
  build_tree (* init_board *) a

let init_tree =
  match load with
    | None -> make_leaf init_board
    | Some f -> 
	printf "Reading dump file %s...@." f;
	let t = read_tree f in
	printf "Done.@.";
	t


let () =
  let b = ref init_tree in
  let time = ref (Unix.time ()) in
  report_iteration !b;
  while true do
    b := expand_tree init_board !b;
    let n = size !b in
    report_iteration !b;
    let t = Unix.time () in
    if t -. !time > 3600.0 (* 60 minutes *) then
      begin
	let f = "dump" ^ (string_of_int n) in
	dump_tree (Filename.concat sol_dir f) !b;
	printf "saved dump file %s@." f;
	time := t;
      end;
  done
