
open Format
open Graphics
open Puzzle
open Board

let graphics = ref false
let resort = ref false
let load = ref None
let merge_files = ref []
let sol_dir = ref "best8"
let credit = ref 5
let first_tile = ref 0
let first_rot = ref (-1)
let stats = ref (-1)
let server_name = ref "localhost"

let spec = 
  ["-p", Arg.String (fun _ -> ()), "<file>  set puzzle file";
   "-credit", Arg.Set_int credit, "<n>  credit for leaves";
   "-load", Arg.String (fun f -> load := Some f), "<file>  load state";
   "-merge", Arg.String (fun f -> merge_files := f :: !merge_files), "<file>  merge state";
   "-graphics", Arg.Set graphics, "with graphical display";
   "-resort", Arg.Set resort, "resorts the tree read from file";
   "-firsttile", Arg.Set_int first_tile, "<n>  search only with this tile first" ;
   "-firstrot", Arg.Set_int first_rot, "<n>  search with this rotation first" ;
   "-stats", Arg.Set_int stats, "<depth> display statistics";
   "-server", Arg.Set_string server_name, "<name>  server name";
   "-dir", Arg.Set_string sol_dir, "<dir>  directory for solutions";]
let () = Arg.parse spec (fun _ -> ()) "brute8.opt [options]"
let graphics = !graphics
let credit = !credit
let resort = !resort
let first_tile = !first_tile
let first_rot = !first_rot
let stats = !stats
let load = !load
let merge_files = !merge_files
let sol_dir = !sol_dir
let server_name = !server_name

let step = 50
let half = step / 2
let scale = half-1

let () =
  if graphics then Draw.open_board step

let draw_tile i j t =
  let x = step + j * step + half in
  let y = step + i * step + half in
  Draw.piece x y scale t

let clear_tile i j =
  let x = step + j * step in
  let y = step + i * step in
  set_color white;
  fill_rect (x+1) (y+1) (step-2) (step-2)

let border_tile c i j =
  let x = step + j * step in
  let y = step + i * step in
  set_line_width 3;
  set_color c;
  draw_rect (x+2) (y+2) (step-4) (step-4)

let () = 
  if graphics then begin
    for i = 0 to n-1 do
      for j = 0 to n-1 do
	draw_tile i j pieces.(n*i+j)
      done
    done;
    ignore (read_key ());
    for i = 0 to n-1 do
      for j = 0 to n-1 do
	clear_tile i j
      done
    done;
    if is_eternity_ii then 
      begin
	draw_tile 7 7 (piece_at initial 7 7);
	draw_tile 2 2 (piece_at initial 2 2);
      end
  end

let best = ref 0


module MinCard = struct
  (* choose a piece with minimal possibilities cardinality *)
  let choose_in f b =
    let best_c = ref max_int in
    let best = ref (-1,0,Pset.empty) in
    let test i j =
      if f i j then
	let ps = poss b i j in
	let c = Pset.cardinal ps in
	if c < !best_c then begin
	  best_c := c;
	  best := (i,j,ps)
	end
    in
    iter_empty test b;
    match !best with
      | (-1,_,_) -> raise Not_found
      | s -> s

  (* choose a piece with minimal distance to corners *)
  let choose_close_to_corners f b =
    let best_c = ref max_int in
    let best = ref (-1,0,Pset.empty) in
    let test i j =
      if f i j then
	let ps = poss b i j in
	let c = Pset.cardinal ps in
	let c = 5*c + 3*corner_distance i j in
	if c < !best_c then begin
	  best_c := c;
	  best := (i,j,ps)
	end
    in
    iter_empty test b;
    match !best with
      | (-1,_,_) -> raise Not_found
      | s -> s

  (* choose a piece with minimal possibilities cardinality *)
  let choose b =
    let best_c = ref max_int in
    let best = ref (-1,0,Pset.empty) in
    let test i j =
      let ps = poss b i j in
      let c = Pset.cardinal ps in
      if c < !best_c then begin
	best_c := c;
	best := (i,j,ps)
      end
    in
    iter_empty test b;
    match !best with
      | (-1,_,_) -> assert false
      | s -> s

  (* choose a piece with minimal possibilities cardinality
     apart from the 3x3 region around (7,7) *)
  let choose_apart77 = 
    choose_in (fun i j -> i < 6 || i > 8 || j < 6 || j > 8)

  let choose3 b =
    if todo b < 20 then choose b else choose_apart77 b

  let avoid = [3,3; 3,11; 11,3; 11,11; 8,8]

  let choose4 b =
    (*try choose_in (fun i j -> j mod 8 = 7) b*)
    (*try choose_in (fun i j -> j <= 8 && i <= 8) b*)
    (*try choose_in (fun i j -> abs (i-7)<=4 && abs(j-7)<=2) b*)
    (*try 
      if (todo b) mod 10 <= 5 then 
	choose_in (fun i j -> i+j <= 12) b
      else
	choose_in (fun i j -> i+j >= 14) b*)
    (*try choose_in (fun i j -> abs (i-7) + abs(j-7) <= 3) b*)
    (*try choose_in (fun i j -> i <= 1 || i >= 14 || j <= 1 || j >= 14) b*)
    (*try choose_in (fun i j -> i <= 0 || i >= 15 || j <= 0 || j >= 15) b*)
    (*try choose_in (fun i j -> not (List.mem (i,j) avoid)) b*)
    (*try choose_in (fun i j -> abs (i-7)>2 || abs(j-7)>2) b*)
    try choose_in (fun i j -> i=0 || i=15 || j=0 || j=15) b
    with Not_found ->
    try choose_in (fun i j -> i=1 || i=14 || j=1 || j=14) b
    (*try choose_in (fun i j -> not (List.mem (i,j) avoid)) b*)
    (*try choose_in (fun i j -> abs (i-7)>2 || abs(j-7)>2) b*)
    with Not_found -> choose b

  let connected b i j =
    (i > 0 && is_real_piece (piece_at b (i-1) j)) ||
      (i < n-1 && is_real_piece (piece_at b (i+1) j)) ||
      (j > 0 && is_real_piece (piece_at b i (j-1))) ||
      (j < n-1 && is_real_piece (piece_at b i (j+1))) 

  let choose_connected b =
    choose_in (connected b) b

  let choose_connected_close_to_corners b =
    choose_close_to_corners (connected b) b

  let choose_connected_ondiag_if_possible b =
    try 
      choose_in 
	(fun i j -> 
	   0 <= j - i && j - i <= 1 && connected b i j) b
    with
	Not_found -> choose_in (connected b) b
    
end

let dump b score =
  let file = Filename.concat sol_dir (string_of_int score ^ ".sol") in
  if is_eternity_ii && not (Sys.file_exists file) then begin
    Board.save b file;
    printf "saved in %s.@." file
  end

let nsol = ref 0

let solution b =
  incr nsol;
  printf "solution %d!@." !nsol;
  dump b 500;
  read_key ()


let place ~complete b i j p =
  let b = Board.place ~complete b i j p in
  let score = score b in
  if score > !best then begin
    printf "score = %d@." score;
    best := score;
    if score > 280 then dump b score
  end;
  b

let count = ref 0


let isqrt i =
  if i <= 0 then 0
  else if i <= 1 then 1 
  else if i <= 4 then 2 
  else if i <= 9 then 3
  else if i <= 16 then 4
  else if i <= 25 then 5
  else if i <= 36 then 6
  else if i <= 49 then 7
  else if i <= 64 then 8
  else if i <= 81 then 9
  else if i <= 100 then 10
  else if i <= 121 then 11
  else if i <= 144 then 12
  else if i <= 169 then 13
  else 14

module IntMap = 
  Map.Make(struct
	     type t = int
	     let compare i1 i2 = Pervasives.compare i2 i1
	   end)

let decr_avail avail p =    
  let avail_p = Parray.get avail p - 1 in
  Parray.set avail p avail_p

(* variance of an array

  sum = sum_{i}  t[i]

  moy = sum / n

  var = sum_{i} (t[i]-moy)^2
      = sum_{i} (t[i]^2 - 2 * t[i] * moy + moy ^ 2)
      = sum_{i} t[i]^2 - 2 * moy * sum + n * moy ^ 2 
      = sum_{i} t[i]^2 - 2 * sum^2 / n  + sum ^ 2 / n 
      = sum_{i} t[i]^2 - sum^2 / n   

*)
let variance avail : int =
  let sum = ref 0 in
  let squares = ref 0 in
  Parray.iter
    (fun v ->
       sum := !sum + v;
       squares := !squares + v * v) avail;
  23 * !squares - !sum * !sum

let rec loop credit b pv = 
  if todo b = 0 then (solution b; raise Incorrect);
  (*incr count; if !count = 50 then begin count := 0; check b end;*)
  let i,j,ps = 
    if todo b = n2
    then MinCard.choose b 
    else MinCard.choose_connected_ondiag_if_possible b 
  in
(* heuristics 1: maximizes remaining possibilities *)
(*
  let t = ref IntMap.empty in
  let () = Pset.iter
    (fun p ->
       try 
	 let b = place ~complete:true b i j p in 
	 let possibilities = ref 0 in
	 iter_empty
	   (fun i j -> possibilities := !possibilities + 
	      isqrt (Pset.cardinal (poss b i j))) b;
	 let x =
	   try
	     IntMap.find !possibilities !t
	   with Not_found -> Pset.empty
	 in
	 t := IntMap.add !possibilities (Pset.add p x) !t 
       with Incorrect | Incomplete -> ())
    ps 
  in
  let c = ref credit in
  let sc = Board.score b in
  let s = ref sc in
  begin
    try
      IntMap.iter
	(fun _ ps -> 
	   Pset.iter 
	     (fun p -> 
		try 
		  let cr = !c in
		  decr c;
		  if !c < 0 then raise Exit;
		  let b1 = place ~complete:true b i j p in 
		  if graphics then draw_tile i j p;
		  let cr = 
		    let sc1 = Board.score b1 in
		    if sc1 >= 150 && sc / 50 < sc1 / 50 then cr+1 else cr
		  in
		  let sc = loop cr b1 in
		  if sc > !s then s := sc;
		  if graphics then clear_tile i j
		with Incorrect | Incomplete -> ())
	     ps)
	!t;
    with Exit -> ();
  end;
*)
(* heuristics 2: minimizes variance of remaining tiles *)
(*
  let avail = Board.avail b in
  let l = Pset.fold
    (fun p acc ->
       let a = decr_avail avail p.up in
       let a = decr_avail a p.left in
       let a = decr_avail a p.down in
       let a = decr_avail a p.right in
       let var = variance a in
       (var,p)::acc)
    ps []
  in
  let l = List.sort (fun (var1,p1) (var2,p2) -> Pervasives.compare var1 var2) l in
*)
  let c = ref credit in
  let s = ref (Board.score b) in
  let pvr = ref [] in
  let one_iter pvrem p =
    try 
      if !c <= 0 then raise Exit;
      let b1 = place ~complete:true b i j p in 
      if graphics then draw_tile i j p;
      let sc,pv1 = loop (!c-1) b1 pvrem in
      if sc > !s then (s := sc; pvr := (i,j,p)::pv1);
      c := !c - 48;
      if graphics then clear_tile i j	       
    with Incorrect | Incomplete -> ()    
  in
  begin
    try
      match pv with
	| [] -> Pset.random_iter (one_iter []) ps
	| (i,j,p)::rem ->
	    one_iter rem p;
	    Pset.random_iter (one_iter []) (Pset.remove p ps)
    with Exit -> ()
  end;
  (* returns score and pv *)
  !s, !pvr

let iter_grid f =
  for i = 0 to n-1 do for j = 0 to n-1 do f i j done done

let init_board = Board.initial

type search_tree =
  | Unsolvable
  | Leaf of int * int * (int * int * Puzzle.piece) list 
      (* score, credit, principal variation *)
  | Node of int * int * int * int * (Puzzle.piece * search_tree) list
      (* score, i, j, size, children *)

let score = function
  | Unsolvable -> 0
  | Leaf(s,_,_) -> s
  | Node(s,_,_,_,_) -> s

let size = function
  | Unsolvable -> max_int
  | Leaf _ -> 0
  | Node(_,_,_,s,_) -> s

(* make a leaf with credit *)
let make_leaf b pv =
(*
  let sc = Board.score b in
*)
  let c = 
    (*
      if sc >= 150 then credit + sc / 50 - 2 else credit 
    *)
    256
  in
  let s,pv = loop c b pv in
  Leaf(s,c,pv)

let mesure s siz =
  10.0 *. log (float (s+1)) -. log (float (siz+1))
 
let rec insert p t s siz l =
  match l with
    | [] -> [(p,t)]
    | (p',t')::r ->
	let s' = score t' in
	let siz' = size t' in
	if mesure s siz >= mesure s' siz'
	then (p,t)::l
	else (p',t')::(insert p t s siz r)
	
let rec expand_tree b t =
  match t with
    | Unsolvable -> assert false
    | Leaf(s,c,pv) ->
	let i,j,ps = 
	  if todo b = n2
	  then MinCard.choose b 
	  else MinCard.choose_connected_ondiag_if_possible b 
	in
	let l = Pset.fold
	  (fun p acc ->
	     try 
	       let b = place ~complete:true b i j p in 
	       if graphics then (draw_tile i j p; border_tile red i j); 
	       let l = make_leaf b pv in
	       if graphics then clear_tile i j;
	       (p,l)::acc
	     with Incorrect | Incomplete -> acc)
	  ps []
	in
	if l = [] then Unsolvable else
	  let l = List.sort 
	    (fun (p1,t1) (p2,t2) -> Pervasives.compare (score t2) (score t1)) l 
	  in
	  Node(score (snd (List.hd l)), i, j, 1, l)
    | Node(s,i,j,_,[]) -> assert false
    | Node(s,i,j,siz,(p,t)::l) -> 
	let b = place ~complete:true b i j p in 
	if graphics then (draw_tile i j p; border_tile blue i j);
	let t = expand_tree b t in
	if graphics then clear_tile i j;
	let sc = 
	  List.fold_left
	    (fun acc (_,t) -> max acc (score t))
	    (score t)
	    l
	in
	let siz' = size t in
	Node(sc,i,j,siz+1,insert p t (score t) siz' l)
	
let rec put_first_tile_first first_tile first_rot acc l =
  match l with
    | [] -> raise Not_found
    | ((p,_) as t)::r ->
	if p.id = first_tile && p.rot = first_rot then 
	  t::List.rev_append acc r 
	else
	  put_first_tile_first first_tile first_rot (t::acc) r

let expand_tree_root b first_tile first_rot t =
  match t with
    | Unsolvable -> assert false
    | Leaf _ -> assert false
    | Node (_,_,_,_,[]) -> assert false
    | Node(s,i,j,siz,(((p,t)::r) as l)) ->
	if p.id <> first_tile || p.rot <> first_rot then
	  try
	    Node(s,i,j,siz,put_first_tile_first first_tile first_rot [] l)
	  with
	      Not_found ->
		let l = List.sort
		  (fun (p,_) (q,_) -> Pervasives.compare p.id q.id) l
		in
		eprintf "First tile/rot must be among@\n@[";
		List.iter (fun (p,_) -> eprintf "%d,%d@ " p.id p.rot) l;
		eprintf "@]@.";
		raise Exit
	else
	  let b = place ~complete:true b i j p in 
	  if graphics then (draw_tile i j p; border_tile green i j);
	  let t = expand_tree b t in
	  if graphics then clear_tile i j;
	  let sc = 
	    List.fold_left
	      (fun acc (_,t) -> max acc (score t))
	      (score t)
	      r
	  in
	  Node(sc,i,j,siz+1,(p,t)::r)
	       
	
let dump_tree file t =
  let ch = open_out file in
  let rec loop fmt t =
    match t with
      | Unsolvable -> fprintf fmt "U"
      | Leaf(s,c,pv) -> 
	  fprintf fmt "@[L(%d,%d,@,@[[" s c;
	  List.iter
	    (fun (i,j,p) -> fprintf fmt "(%d,@,%d,@,%d,@,%d);@ " i j p.id p.rot)
	    pv;	  
	  fprintf fmt "]@])@]"
      | Node(s,i,j,siz,l) ->
	  fprintf fmt "@[N(%d,%d,%d,%d,@,@[[" s i j siz;
	  List.iter
	    (fun (p,t) -> fprintf fmt "(%d,@,%d,@,%a);@ " p.id p.rot loop t)
	    l;
	  fprintf fmt "]@])@]"
  in
  loop (formatter_of_out_channel ch) t;
  close_out ch

let tree_to_ast t =
  let rec loop t =
    match t with
      | Unsolvable -> Ast.U
      | Leaf(s,c,pv) -> 
	  Ast.L(s,c,List.map (fun (i,j,p) -> (i,j,p.id,p.rot)) pv)
      | Node(s,i,j,siz,l) ->
	  let l = 
	    List.map 
	      (fun (p,t) -> (p.id,p.rot,loop t))
	      l	  
	  in Ast.N(s,i,j,siz,l)
  in
  loop t

let rec build_tree resort a =
  match a with
    | Ast.U -> Unsolvable
    | Ast.L(score,credit,l) -> 
	let pv =
	  List.map
	    (fun (i,j,id,rot) -> 
	       let p = Puzzle.get_piece id rot in
	       (i,j,p))
	    l	       
	in
	Leaf(score,credit,pv)
    | Ast.N(scor,i,j,siz,l) ->
	let (sc,si,l) =
	  List.fold_right
	    (fun (p,r,a) (sc,siz,l) ->	       
	       let p = Puzzle.get_piece p r in
	       (* let b = place ~complete:true b i j p in *)
	       let t = build_tree resort a in
	       let sco = score t in
	       let si = size t in
	       (max sc sco,siz+si,
		if resort then insert p t sco si l else (p,t)::l))
	    l (0,1,[])
	in
(*
	if sc <> scor then 
	    failwith "bad score" 
	else
	  if si <> siz then 
	    begin
	      printf "bad size: expected %d, got %d@." si siz;
	      failwith "bad size"
	    end
	  else
	 *)
	Node(sc,i,j,si,l)

let build_tree_root first_tile first_rot resort a =
  match a with
    | Ast.U -> Unsolvable
    | Ast.L(score,credit,l) -> 
	let pv =
	  List.map
	    (fun (i,j,id,rot) -> 
	       let p = Puzzle.get_piece id rot in
	       (i,j,p))
	    l	       
	in
	Leaf(score,credit,pv)
    | Ast.N(sco,i,j,siz,l) ->
	let l =
	  List.fold_right
	    (fun (p,r,a) l ->	       
	       if p <> first_tile || r <> first_rot then l
	       else
		 let p = Puzzle.get_piece p r in
		 let t = build_tree resort a in
		 (p,t)::l)
	    l []
	in
	Node(sco,i,j,siz,l)

let report_iteration b =	       
  let r = 
    "iteration=" ^ (string_of_int (size b)) ^ 
    " best=" ^ (string_of_int (score b))
  in
  if graphics then 
    set_window_title 
      ("Eternity II, brute8: " ^ r)
  else
    printf "%s@." r

let read_tree file =
  let ch = open_in file in
  let b = Lexing.from_channel ch in
  try
    let a = Parse_tree.tree Lex_tree.token b in
    if first_tile > 0 then
      build_tree_root first_tile first_rot resort a
    else
      build_tree resort a
  with Parsing.Parse_error ->
    eprintf "parse error at char %d@." b.Lexing.lex_curr_p.Lexing.pos_cnum;
    exit 1


let init_tree () =
  match load with
    | None -> 
	(*
	  eprintf "option -load is mandatory@.";
	  assert false 
	*)
	make_leaf init_board []
    | Some f -> 
	printf "Reading dump file %s...@." f;
	let t = read_tree f in
	printf "Done.@.";
	t

let nul_size_in_list l =
  List.exists (fun (_,t) -> size t = 0) l

let rec display_stats ~prof b =
  match b with
    | Node(_,_,_,_,l) ->
	printf "@[<v 4>    ";
	List.iter
	  (fun (p,t) ->
	     printf "%3d,%d: " p.id p.rot;
	     let s = score t and siz = size t in
	     printf "score=%d size=%d mesure=%f@\n" s siz (mesure s siz);
	     if prof > 0 then display_stats ~prof:(prof-1) t else
	     match t with 
	       | Node(_,_,_,_,l) ->
		   printf "    ";
		   List.iter
		     (fun (_,t) ->
			let s = (size t) / 2 in
			if s<10 then printf "%c" (Char.chr (s+Char.code '0')) 
			else 
			if s<36 then printf "%c" (Char.chr (s+Char.code 'A'-10)) 
			else 
			if s<62 then printf "%c" (Char.chr (s+Char.code 'a'-36)) else printf "+")
		     l;
		   printf "@\n"
		   (*
		   if nul_size_in_list l then 
		     printf "    *nul size remaining*@\n"
		   else
		     if List.exists 
		       (fun (_,t) -> 
			  match t with
			    | Node(_,_,_,_,l) -> nul_size_in_list l
			    | _ -> assert false)
		       l 
		     then 
		       printf "    *nul size remaining at level 1*@\n"
		   *)
	       | Unsolvable -> printf "    *unsolvable*@\n"
	       | Leaf _ -> printf "    *leaf*@\n")
	  l;
	printf "@]@\n";
    | _ -> ()

let compare_tiles (p1,_) (p2,_) =
  match Pervasives.compare p1.id p2.id with
    | 0 -> Pervasives.compare p1.rot p2.rot
    | c -> c

let rec merge_sons (l1 : (Puzzle.piece * search_tree) list) l2 =
  match l1,l2 with
    | [], _ -> l2
    | _, [] -> l1
    | (((p1,t1) as e1)::r1),(((p2,t2) as e2)::r2) ->
	let c = compare_tiles e1 e2 in
	if c < 0 then e1::(merge_sons r1 l2) else
	if c > 0 then e2::(merge_sons l1 r2) else
	(p1,merge t1 t2)::(merge_sons r1 r2)
	
and merge t1 t2 =
  match t1,t2 with
    | Unsolvable,_ 
    | _,Unsolvable -> Unsolvable
    | Leaf _,t
    | t,Leaf _ -> t
    | Node(sc1,i1,j1,siz1,l1),Node(sc2,i2,j2,siz2,l2) ->
	if i1<>i2 || j1 <> j2 then
	  failwith "merge impossible: not the same free cell selected"
	else
	  let l1 = List.sort compare_tiles l1 in
	  let l2 = List.sort compare_tiles l2 in
	  let l = merge_sons l1 l2 in
(*
		    List.map2
	    (fun (p1,t1) (p2,t2) ->
	       if p1 != p2 then
		 begin
		   failwith "merge impossible: not the same sons ???"
		 end
	       else
		 (p1,merge t1 t2))
	    l1 l2
	  in
*)
	  let (sc,siz) =
	    List.fold_right
	      (fun (p,t) (sc,siz) ->	       
		 (max sc (score t),siz+(size t)))
	      l (0,1)
	  in
	  Node(sc,i1,j1,siz,l)
	

(* main launch *)
  
let () =
  let b = ref (init_tree ())in
  List.iter
    (fun f -> 
       printf "Reading dump file %s...@." f;
       let t = read_tree f in
       printf "Merging...@.";
       b := merge !b t;
       printf "Done.@.")
    merge_files;
  if merge_files <> [] then
    begin
      let f = "merged_" ^ (string_of_int (size !b)) in
      dump_tree (Filename.concat sol_dir f) !b;
      printf "saved merged dump file %s@." f;
    end ;
  let time = ref (Unix.time ()) in
  report_iteration !b;
  if stats >= 0 then begin display_stats ~prof:stats !b; printf "@." end
  else
    while true do
      b := 
	if first_tile <> 0 then
	  expand_tree_root init_board first_tile first_rot !b
	else
	  expand_tree init_board !b;
      let n = size !b in
      report_iteration !b;
      let t = Unix.time () in
      if t -. !time > 36000.0 (* 10 hours *) then
	begin
	  let f = "dump_" ^ (string_of_int first_tile) ^ "_" 
	    ^ (string_of_int first_rot) ^ "_" ^(string_of_int n) in
	  dump_tree (Filename.concat sol_dir f) !b;
	  printf "saved dump file %s@." f;
	  time := t;
	end;
    done

