
open Format
open Graphics

type point = float * float (* r, theta(radians) *)

type command =
  | FillPoly of color * point list
  | FillArc of
      color * point(*center*) * float(*radius*) * int(*a1(d)*) * int(*a2(d)*)

type drawing = command list

(* cl: XY command list *)
let draw x0 y0 scale cl =
  let scale = float scale in
  let tr (x,y) = 
    x0 + truncate (scale *. x), y0 + truncate (scale *. y) 
  in
  List.iter
    (function
       | FillPoly (c, pl) ->
	   set_color c;
	   fill_poly (Array.of_list (List.map tr pl))
       | FillArc (c, p, r, a1, a2) ->
	   set_color c;
	   let x,y = tr p in
	   let r = truncate (scale *. r) in
	   fill_arc x y r r a1 a2)
    cl

let pi = 4. *. atan 1.
let pi_2 = pi /. 2.
let _pi_2 = -. pi_2
let pi_4 = pi /. 4.
let _pi_4 = -. pi_4

let xy (r,th) = r *. cos th, r *. sin th
let rth (x,y) = sqrt (x *. x +. y *. y), atan2 y x

let map fp fa = function
  | FillPoly (c, pl) -> FillPoly (c, List.map fp pl)
  | FillArc (c, p, r, a1, a2) -> FillArc (c, fp p, r, fa a1, fa a2)

let rotate a (r, th) = (r, th +. a)

let translate (dx,dy) (x,y) = x+.dx, y+.dy

let tr_right = 
  map (fun p -> translate (1.,0.) (xy (rotate pi p))) (fun a -> a+180)
let tr_up = 
  map (fun p -> translate (0.,1.) (xy (rotate _pi_2 p))) (fun a -> a-90)
let tr_left = 
  map (fun p -> translate (-1.,0.) (xy p)) (fun a -> a)
let tr_down =
  map (fun p -> translate (0.,-1.) (xy (rotate pi_2 p))) (fun a -> a+90)

let draw_right x y s cl = draw x y s (List.map tr_right cl)
let draw_up x y s cl = draw x y s (List.map tr_up cl)
let draw_left x y s cl = draw x y s (List.map tr_left cl)
let draw_down x y s cl = draw x y s (List.map tr_down cl)

(* patterns *)

let background c = FillPoly (c, [1.,_pi_2; 1.,0.; 1.,pi_2])

let marine = rgb 46 46 71	    
let orange = rgb 252 131 1
let ciel = rgb 107 197 190
let rose = rgb 253 78 115
let vert = rgb 47 167 27
let jaune = rgb 246 254 14
let violet = rgb 99 20 108
let marron = rgb 117 8 26

let fleur = 
  [ background marine; 
    FillArc (yellow, (0.33,0.), 0.33, 0, 360);
    FillArc (yellow, (0.33,pi_2), 0.33, -90, 90);
    FillArc (yellow, (0.33,_pi_2), 0.33, -90, 90);
    FillArc (marine, (0.,0.), 0.33, -90, 90) ]
   
let shuriken =
  [ background orange;
    FillArc (ciel, (0.,0.), 0.6, -90, 90);
    FillArc (orange, (0.7,pi_4), 0.3, 180-45, 360-45);
    FillArc (orange, (0.7,_pi_4), 0.3, 45, 180+45)
  ]

let diamand =
  let r = sqrt 2. /. 2. in
  [ background rose;
    FillPoly (ciel, [0.5,_pi_2; r,_pi_4; r,pi_4; 0.5,pi_2]);
    FillArc (rose, (r,pi_4), 0.3, 180-1, 270+1); 
    FillArc (rose, (r,_pi_4), 0.3, 90, 180); 
    FillPoly (rose, [0.25,_pi_2; 0.25,0.; 0.25,pi_2])
  ]

let ecrou =
  let th1 = atan 0.5 in
  let th2 = atan 2.0 in
  let r = sqrt 5. /. 4. in
  [ background vert;
    FillPoly (marine, [0.5,_pi_2; r,-.th2; r,-.th1; r,th1; r,th2; 0.5,pi_2]);
    FillArc (vert, (0.,0.), 0.25, -90, 90);
  ]

let carre_bleu =
  [ background marine;
    FillPoly (ciel, [0.67,_pi_2; 0.67,0.; 0.67,pi_2]);
    FillPoly (marine, [0.33,_pi_2; 0.33,0.; 0.33,pi_2]);
  ]

let carre_vert =
  [ background jaune;
    FillPoly (vert, [0.67,_pi_2; 0.67,0.; 0.67,pi_2]);
    FillPoly (jaune, [0.33,_pi_2; 0.33,0.; 0.33,pi_2]);
  ]

let etoile_bleue =
  let r1 = 0.65 in
  let r2 = 0.35 in
  let pi_3 = pi /. 3. in
  let pi_6 = pi /. 6. in
  [ background jaune;
    FillPoly (ciel, [r1,_pi_2; r2,-.pi_3; r1,-.pi_6; r2,0.;
		     r1,pi_6; r2,pi_3; r1,pi_2]) ]

let etoile_jaune =
  let r1 = 0.65 in
  let r2 = 0.35 in
  let pi_3 = pi /. 3. in
  let pi_6 = pi /. 6. in
  [ background marron;
    FillPoly (jaune, [r1,_pi_2; r2,-.pi_3; r1,-.pi_6; r2,0.;
		     r1,pi_6; r2,pi_3; r1,pi_2]) ]

let etoile_violette =
  let r1 = 0.65 in
  let r2 = 0.35 in
  let pi_3 = pi /. 3. in
  let pi_6 = pi /. 6. in
  [ background orange;
    FillPoly (violet, [r1,_pi_2; r2,-.pi_3; r1,-.pi_6; r2,0.;
		     r1,pi_6; r2,pi_3; r1,pi_2]) ]

let croix_violette =
  let th1 = atan 0.5 in
  let th2 = atan 2.0 in
  let d = 0.2 in
  let r = sqrt 5. *. d in
  [ background violet;
    FillArc (jaune, (0.,0.), 0.6, -90, 90);
    FillPoly (violet, [d,pi_2; r,th2; r,th1; d,0.; r,-.th1; r,-.th2; d,_pi_2])]

let croix_bleue =
  let th1 = atan 0.5 in
  let th2 = atan 2.0 in
  let d = 0.2 in
  let r = sqrt 5. *. d in
  [ background marine;
    FillArc (rose, (0.,0.), 0.6, -90, 90);
    FillPoly (marine, [d,pi_2; r,th2; r,th1; d,0.; r,-.th1; r,-.th2; d,_pi_2])]

let croix_marron =
  let th1 = atan 0.5 in
  let th2 = atan 2.0 in
  let d = 0.2 in
  let r = sqrt 5. *. d in
  [ background marron;
    FillArc (vert, (0.,0.), 0.6, -90, 90);
    FillPoly (marron, [d,pi_2; r,th2; r,th1; d,0.; r,-.th1; r,-.th2; d,_pi_2])]

let trefle_rose =
  let th1 = atan 0.3333 in
  let th2 = atan 3.0 in
  let d = 0.15 in
  let r = sqrt 10. *. d in
  [ background vert;
    FillArc (rose, (0.5,0.), 0.25, 0, 360);
    FillPoly (rose, [3.*.d,pi_2; r,th2; sqrt 2. *. d,pi_4; r,th1;
		     r,-.th1; sqrt 2. *. d,_pi_4; r,-.th2; 3.*.d,_pi_2]);
    FillArc (rose, (0.5,pi_2), 0.25, -90, 90);
    FillArc (rose, (0.5,_pi_2), 0.25, -90, 90) ]
    
let trefle_jaune =
  let th1 = atan 0.3333 in
  let th2 = atan 3.0 in
  let d = 0.15 in
  let r = sqrt 10. *. d in
  [ background rose;
    FillArc (jaune, (0.5,0.), 0.25, 0, 360);
    FillPoly (jaune, [3.*.d,pi_2; r,th2; sqrt 2. *. d,pi_4; r,th1;
		     r,-.th1; sqrt 2. *. d,_pi_4; r,-.th2; 3.*.d,_pi_2]);
    FillArc (jaune, (0.5,pi_2), 0.25, -90, 90);
    FillArc (jaune, (0.5,_pi_2), 0.25, -90, 90) ]
    
let trefle_orange =
  let th1 = atan 0.3333 in
  let th2 = atan 3.0 in
  let d = 0.15 in
  let r = sqrt 10. *. d in
  [ background marine;
    FillArc (orange, (0.5,0.), 0.25, 0, 360);
    FillPoly (orange, [3.*.d,pi_2; r,th2; sqrt 2. *. d,pi_4; r,th1;
		       r,-.th1; sqrt 2. *. d,_pi_4; r,-.th2; 3.*.d,_pi_2]);
    FillArc (orange, (0.5,pi_2), 0.25, -90, 90);
    FillArc (orange, (0.5,_pi_2), 0.25, -90, 90) ]

let chateau_jaune =
  let h = 0.125 in
  [ background rose;
    FillPoly (jaune, [0.75,pi_2; rth (0.25,0.5); rth (h,3.*.h);
		      rth (3.*.h,h); rth (0.5,0.25); 0.75,0.;
		      rth (0.5,-0.25); rth (3.*.h,-.h); rth (h,-.3.*.h);
		      rth (0.25,-0.5); 0.75,_pi_2]) ]
    
let chateau_bleu =
  let h = 0.125 in
  [ background jaune;
    FillPoly (marine, [0.75,pi_2; rth (0.25,0.5); rth (h,3.*.h);
		       rth (3.*.h,h); rth (0.5,0.25); 0.75,0.;
		       rth (0.5,-0.25); rth (3.*.h,-.h); rth (h,-.3.*.h);
		       rth (0.25,-0.5); 0.75,_pi_2]) ]
    
let chateau_rose =
  let h = 0.125 in
  [ background ciel;
    FillPoly (rose, [0.75,pi_2; rth (0.25,0.5); rth (h,3.*.h);
		     rth (3.*.h,h); rth (0.5,0.25); 0.75,0.;
		     rth (0.5,-0.25); rth (3.*.h,-.h); rth (h,-.3.*.h);
		     rth (0.25,-0.5); 0.75,_pi_2]) ]

let cristal_rose =
  let d = 0.15 in
  [ background ciel;
    FillPoly (rose, [0.75,pi_2; rth (0.25,0.5); rth (d,d); rth (0.5,0.25);
		     0.75,0.; rth (0.5,-0.25); rth (d,-.d); rth (0.25,-0.5);
		     0.75,_pi_2]) ]
    
let cristal_bleu =
  let d = 0.15 in
  [ background violet;
    FillPoly (ciel, [0.75,pi_2; rth (0.25,0.5); rth (d,d); rth (0.5,0.25);
		     0.75,0.; rth (0.5,-0.25); rth (d,-.d); rth (0.25,-0.5);
		     0.75,_pi_2]) ]
    
let cristal_orange =
  let d = 0.15 in
  [ background vert;
    FillPoly (orange, [0.75,pi_2; rth (0.25,0.5); rth (d,d); rth (0.5,0.25);
		       0.75,0.; rth (0.5,-0.25); rth (d,-.d); rth (0.25,-0.5);
		       0.75,_pi_2]) ]
    
let biscuit = 
  let d = 0.2 in
  [ background marron;
    FillArc (orange, (0.,0.), 0.5, -90, 90);
    FillArc (orange, (0.5,0.), 0.25, 0, 360);
    FillArc (orange, (0.5,pi_2), 0.25, -90, 90);
    FillArc (orange, (0.5,_pi_2), 0.25, -90, 90);
    FillPoly (marron, [0.5,pi_2; rth (d,d); 0.5,0.; rth (d,-.d); 0.5,_pi_2]) ]
    
let patterns = 
  [| (*  0 *) [background (rgb 70 70 70)]; 
     (*  1 *) chateau_jaune; 
     (*  2 *) chateau_bleu; 
     (*  3 *) chateau_rose;
     (*  4 *) etoile_violette; 
     (*  5 *) etoile_bleue; 
     (*  6 *) etoile_jaune;
     (*  7 *) croix_violette; 
     (*  8 *) croix_bleue; 
     (*  9 *) croix_marron;
     (* 10 *) carre_bleu;
     (* 11 *) carre_vert;
     (* 12 *) biscuit;
     (* 13 *) fleur;
     (* 14 *) trefle_rose;
     (* 15 *) trefle_jaune;
     (* 16 *) trefle_orange;
     (* 17 *) cristal_rose;
     (* 18 *) cristal_bleu;
     (* 19 *) cristal_orange;
     (* 20 *) shuriken;
     (* 21 *) diamand;
     (* 22 *) ecrou; 
  |]
     
let right x y s p = assert (0 <= p && p <= 22); draw_right x y s patterns.(p)
let up x y s p = assert (0 <= p && p <= 22); draw_up x y s patterns.(p)
let left x y s p = assert (0 <= p && p <= 22); draw_left x y s patterns.(p)
let down x y s p = assert (0 <= p && p <= 22); draw_down x y s patterns.(p)

open Puzzle

let piece x y scale t =
  right x y scale t.right;
  up x y scale t.up;
  left x y scale t.left;
  down x y scale t.down

let open_board ?title ?(lines=true) step =
  open_graph (sprintf " %dx%d" ((n+2) * step) ((n+2) * step));
  let t = 
    match title with
      | None -> "Eternity II"
      | Some t -> t
  in
  set_window_title t;
  if lines then begin
    set_color black;
    for i = 0 to n do
      moveto step (step+i*step); lineto (step+n*step) (step+i*step);
      moveto (step+i*step) step; lineto (step+i*step) (step+n*step)
    done
  end
