
(* graphical editor *)

open Format
open Graphics
open Puzzle

let file = ref None
let () = 
  Arg.parse ["-p", Arg.String (fun _ -> ()), ""]
    (fun f -> file := Some f) "editor [-p file.data] file.sol"
let file = match !file with
  | None -> eprintf "editor [-p file.data] file.sol"; exit 1 
  | Some f -> f

let step = 55
let half = step / 2
let scale = half-1

let () = Draw.open_board step

let draw_tile i j t = 
  let x = step + j * step + half in
  let y = step + i * step + half in
  Draw.piece x y scale t

let color_cell i j c =
  let x = step + j * step in
  let y = step + i * step in
  set_color c;
  fill_rect (x+1) (y+1) (step-2) (step-2)

let gris = rgb 150 150 150

let init_sol = Array.create_matrix n n dummy_piece

let sol = load file

let score = ref 0
let best = ref 0

let dump score =
  let file = Filename.concat "edit" (string_of_int score ^ ".sol") in
  if is_eternity_ii && not (Sys.file_exists file) then begin
    save file sol;
    printf "saved.@."
  end

(* used pieces *)

module Used = 
  Set.Make
    (struct 
      type t = piece 
      let compare p1 p2 = Pervasives.compare p1.id p2.id 
    end)

let used = ref Used.empty

(* tries to place t at (i,j) and returns true if possible *)

let compat i j p =
  (i = 0 || sol.(i-1).(j) == dummy_piece || sol.(i-1).(j).up = p.down) &&
  (i = n-1 || sol.(i+1).(j) == dummy_piece || sol.(i+1).(j).down = p.up) && 
  (j = 0 || sol.(i).(j-1) == dummy_piece || sol.(i).(j-1).right = p.left) &&
  (j = n-1 || sol.(i).(j+1) == dummy_piece || sol.(i).(j+1).left = p.right)

let place i j p =
  compat i j p &&
  begin
    sol.(i).(j) <- p;
    used := Used.add p !used;
    (* score update *)
    if i > 0 && is_real_piece sol.(i-1).(j) then incr score;
    if i < n-1 && is_real_piece sol.(i+1).(j) then incr score;
    if j > 0 && is_real_piece sol.(i).(j-1) then incr score;
    if j < n-1 && is_real_piece sol.(i).(j+1) then incr score;
    draw_tile i j p;
    true
  end 
    
let remove i j =
  let p = sol.(i).(j) in
  sol.(i).(j) <- dummy_piece;
  used := Used.remove p !used;
  (* score update *)
  if i > 0 && is_real_piece sol.(i-1).(j) then decr score;
  if i < n-1 && is_real_piece sol.(i+1).(j) then decr score;
  if j > 0 && is_real_piece sol.(i).(j-1) then decr score;
  if j < n-1 && is_real_piece sol.(i).(j+1) then decr score;
  color_cell i j white

let iter_grid f =
  for i = 0 to n-1 do for j = 0 to n-1 do f i j done done

let iter_empty f =
  iter_grid (fun i j -> if sol.(i).(j) == dummy_piece then f i j)

let () = 
  iter_grid 
    (fun i j -> 
      let p = init_sol.(i).(j) in 
      if is_real_piece p then assert (place i j p))

module Todo = 
  Set.Make(struct type t = int * int
		  let compare = Pervasives.compare end)

let todo = ref Todo.empty

module Poss = struct

  module Pset = struct
    include 
      Set.Make
      (struct 
	 type t = piece
	 let compare p1 p2 = Pervasives.compare (p1.id,p1.rot) (p2.id,p2.rot)
       end)
      
    let of_array a =
      Array.fold_left (fun s p -> add p s) empty a
  end

  let poss = 
    Array.init n 
      (fun i -> 
	 Array.init n 
	   (fun j ->
	      if i = 0 && j = 0 then
		Pset.of_array corners_rot1
	      else if i = 0 && j = n-1 then
		Pset.of_array corners_rot2
	      else if i = n-1 && j = 0 then
		Pset.of_array corners_rot0
	      else if i = n-1 && j = n-1 then
		Pset.of_array corners_rot3
	      else if i = 0 then
		Pset.of_array edges_rot2
	      else if i = n-1 then
		Pset.of_array edges_rot0
	      else if j = 0 then
		Pset.of_array edges_rot1
	      else if j = n-1 then
		Pset.of_array edges_rot3
	      else
		Pset.union 
		  (Pset.union 
		     (Pset.of_array others_rot0)
		     (Pset.of_array others_rot1))
		  (Pset.union 
		     (Pset.of_array others_rot2)
		     (Pset.of_array others_rot3))
	   ))

  let iter i j f =
    Pset.iter f poss.(i).(j)

end

let print_score () =
  printf "score = %d@." !score;
  if !score > !best then begin
    best := !score;
    if !score > 200 then dump !score
  end

let nbposs = Array.create_matrix n n 0

let display_poss () =
  let pot = ref 0 in
  iter_empty
    (fun i j ->
      color_cell i j white;
      let c = ref 0 in
      Poss.iter i j 
	(fun p -> if not (Used.mem p !used) && compat i j p then incr c);
      if !c = 0 then color_cell i j gris;
      set_color blue;
      moveto (step + j*step + 5) (step + i*step + 5);
      draw_string (string_of_int !c);
      nbposs.(i).(j) <- !c);
  (* maximal possible score *)
  iter_empty
    (fun i j ->
      if nbposs.(i).(j) > 0 then begin
	if i > 0 then 
 	  if sol.(i-1).(j) == dummy_piece then 
	    (if nbposs.(i-1).(j) > 0 then incr pot)
	  else pot := !pot+2;
	if i < n-1 then
 	  if sol.(i+1).(j) == dummy_piece then 
	    (if nbposs.(i+1).(j) > 0 then incr pot)
	  else pot := !pot+2;
	if j > 0 then 
 	  if sol.(i).(j-1) == dummy_piece then 
	    (if nbposs.(i).(j-1) > 0 then incr pot)
	  else pot := !pot+2;
	if j < n-1 then
 	  if sol.(i).(j+1) == dummy_piece then 
	    (if nbposs.(i).(j+1) > 0 then incr pot)
	  else pot := !pot+2
      end);
  assert (!pot mod 2 = 0);
  printf "potential score = %d@." (!score + !pot/2)

let clear_todo () =
  Todo.iter (fun (i,j) -> color_cell i j white) !todo;
  todo := Todo.empty;
  display_poss ()

exception Solution

let fill_todo_bt () =
  let rec loop todo =
    if Todo.is_empty todo then raise Solution;
    let i,j as cell = Todo.choose todo in
    let todo = Todo.remove cell todo in
    Poss.iter i j 
      (fun p -> 
	if not (Used.mem p !used) && place i j p then
	  try loop todo with Exit -> remove i j);
    raise Exit
  in
  try 
    loop !todo; 
    assert false
  with 
    | Exit -> 
	printf "no solution!@.";
	Todo.iter (fun (i,j) -> color_cell i j white) !todo;
	todo := Todo.empty;
	display_poss ()
    | Solution ->
	todo := Todo.empty;
	display_poss ();
	print_score ()

let greedy () =
  iter_empty
    (fun i j ->
      try
	Poss.iter i j 
	  (fun p -> 
	    if not (Used.mem p !used) && place i j p then raise Exit)
      with Exit ->
	());
  display_poss ();
  print_score ()

let search_for i j =
  printf "search...@.";
  let r = 
    if j < n-1 && sol.(i).(j+1) != dummy_piece then sol.(i).(j+1).left else -1 
  in
  let l = 
    if j > 0 && sol.(i).(j-1) != dummy_piece then sol.(i).(j-1).right else -1 
  in
  let u = 
    if i < n-1 && sol.(i+1).(j) != dummy_piece then sol.(i+1).(j).down else -1 
  in
  let d = 
    if i > 0 && sol.(i-1).(j) != dummy_piece then sol.(i-1).(j).up else -1 
  in
  let matches p =
    let one_match pr pu pl pd = 
      (r = -1 || pr = r) && (l = -1 || pl = l) && 
      (u = -1 || pu = u) && (d = -1 || pd = d)
    in
    one_match p.right p.up p.left p.down ||
    one_match p.down p.right p.up p.left ||
    one_match p.left p.down p.right p.up ||
    one_match p.up p.left p.down p.right
  in
  iter_grid
    (fun i' j' ->
      if sol.(i').(j') != dummy_piece && matches sol.(i').(j') then begin
	printf "match at %d,%d@." i' j';
	color_cell i' j' red
      end
    )

(* interaction loop *)

let ij st = 
  let x = st.mouse_x in
  let y = st.mouse_y in
  if x >= step && x < step+n*step && y >= step && y < step+n*step then
    (y - step) / step, (x - step) / step
  else
    -1,-1

let () =
  display_poss ();
  while true do
    let st = wait_next_event [Button_down; Key_pressed; Mouse_motion] in
    if st.keypressed then begin match st.key with
      | 'q' -> exit 0
      | 'c' -> clear_todo ()
      | 'b' -> fill_todo_bt ()
      | 'g' -> greedy ()
      | 'p' -> display_poss ()
      | 's' -> save "saved.sol" sol; printf "saved in saved.sol@."
      | 'o' -> let i,j = ij st in if i <> -1 then search_for i j
      | 'r' -> 
	  let i,j = ij st in if i <> -1 then begin
	    remove i j;
	    display_poss ();
	    print_score ()
	  end
      | _ -> ()
    end;
    if st.button then begin
      let i,j as cell = ij st in
      if i >= 0 && sol.(i).(j) == dummy_piece then begin
	todo := Todo.add cell !todo;
	color_cell i j yellow
      end
    end
  done
