
(* graphical editor *)

open Format
open Graphics
open Puzzle
open Board

let file = ref None
let sol_dir = ref "edit2"
let () = 
  Arg.parse ["-p", Arg.String (fun _ -> ()), "";
	     "-dir", Arg.Set_string sol_dir, "<dir>  directory for solutions";]
    (fun f -> file := Some f) "editor [-p file.data] file.sol"
let file = match !file with
  | None -> eprintf "editor [-p file.data] file.sol"; exit 1 
  | Some f -> f
let sol_dir = !sol_dir

let step = 55
let half = step / 2
let scale = half-1

let () = Draw.open_board step

let draw_tile i j t = 
  let x = step + j * step + half in
  let y = step + i * step + half in
  Draw.piece x y scale t

let color_cell i j c =
  let x = step + j * step in
  let y = step + i * step in
  set_color c;
  fill_rect (x+1) (y+1) (step-2) (step-2)

let gris = rgb 150 150 150

let init_board = 
  let init_sol = Puzzle.load file in
  Board.create ~complete:false init_sol

let board = ref init_board

let iter_grid f =
  for i = 0 to n-1 do for j = 0 to n-1 do f i j done done

let display () =
  iter_grid 
    (fun i j -> 
      let p = piece_at !board i j in
      if p != dummy_piece then draw_tile i j p else color_cell i j white)

let best = ref 0

let dump score =
  let file = Filename.concat sol_dir (string_of_int score ^ ".sol") in
  if is_eternity_ii && not (Sys.file_exists file) then begin
    save !board file;
    printf "saved.@."
  end

module Todo = struct 
  module Iset = Set.Make(struct type t = int
				let compare = Pervasives.compare end)
  let empty = Iset.empty
  let is_empty = Iset.is_empty
  let add i j t = Iset.add ((i lsl 4) lor j) t
  let mem i j t = Iset.mem ((i lsl 4) lor j) t
  let remove i j t = Iset.remove ((i lsl 4) lor j) t
  let iter f t = Iset.iter (fun n -> f (n lsr 4) (n land 0b1111)) t
end

let todo = ref Todo.empty

let print_score () =
  let score = score !board in
  printf "score = %d@." score;
  if score > !best then begin
    best := score;
    if score > 200 then dump score
  end

let nbposs = Array.create_matrix n n 0

let display_poss () =
  let pot = ref 0 in
  iter_empty
    (fun i j ->
      color_cell i j white;
      let c = Pset.cardinal (poss !board i j) in
      if c = 0 then color_cell i j gris;
      set_color blue;
      moveto (step + j*step + 5) (step + i*step + 5);
      draw_string (string_of_int c);
      nbposs.(i).(j) <- c)
    !board;
  (* maximal possible score *)
  iter_empty
    (fun i j ->
      if nbposs.(i).(j) > 0 then begin
	let sol = piece_at !board in
	if i > 0 then 
 	  if sol (i-1) (j) == dummy_piece then 
	    (if nbposs.(i-1).(j) > 0 then incr pot)
	  else pot := !pot+2;
	if i < n-1 then
 	  if sol (i+1) (j) == dummy_piece then 
	    (if nbposs.(i+1).(j) > 0 then incr pot)
	  else pot := !pot+2;
	if j > 0 then 
 	  if sol (i) (j-1) == dummy_piece then 
	    (if nbposs.(i).(j-1) > 0 then incr pot)
	  else pot := !pot+2;
	if j < n-1 then
 	  if sol (i) (j+1) == dummy_piece then 
	    (if nbposs.(i).(j+1) > 0 then incr pot)
	  else pot := !pot+2
      end)
    !board;
  assert (!pot mod 2 = 0);
  printf "potential score = %d@." (score !board + !pot/2)

let clear_todo () =
  Todo.iter (fun i j -> color_cell i j white) !todo;
  todo := Todo.empty;
  display_poss ()

let display_todo () =
  Todo.iter (fun i j -> color_cell i j yellow) !todo

(* board stack *)

let stack = ref []

let push () = 
  stack := (!board,!todo) :: !stack

let pop () = match !stack with
  | [] -> 
      ()
  | (b,t) :: st -> 
      stack := st;
      board := b;
      todo := t;
      display ();
      display_poss ();
      display_todo ()

(* backtracking search *)

exception Solution

(* choose a piece with minimal possibilities cardinality *)
module MinCard = struct
  let choose_in todo b =
    let best_c = ref max_int in
    let best = ref (-1,0,Pset.empty) in
    let test i j =
      if Todo.mem i j todo then
	let ps = poss b i j in
	let c = Pset.cardinal ps in
	if c < !best_c then begin
	  best_c := c;
	  best := (i,j,ps)
	end
    in
    iter_empty test b;
    match !best with
      | (-1,_,_) -> raise Not_found
      | s -> s
end

let fill_todo_bt () =
  let rec loop todo b =
    if Todo.is_empty todo then (push (); board := b; raise Solution);
    let i,j,ps = MinCard.choose_in todo b in
    let todo = Todo.remove i j todo in
    Pset.iter
      (fun p -> 
	try
	  let b = place ~complete:false b i j p in
	  try loop todo b with Incorrect | Incomplete -> color_cell i j white
	with Incorrect | Incomplete -> 
	  ())
      ps;
    raise Incomplete
  in
  try 
    loop !todo !board; 
    assert false
  with 
    | Incomplete -> 
	printf "no solution!@.";
	Todo.iter (fun i j -> color_cell i j white) !todo;
	todo := Todo.empty;
	display_poss ()
    | Solution ->
	Todo.iter (fun i j -> draw_tile i j (piece_at !board i j)) !todo;
	todo := Todo.empty;
	display_poss ();
	print_score ()

let greedy () =
  printf "todo...@."
(****
  iter_empty
    (fun i j ->
      try
	Poss.iter i j 
	  (fun p -> 
	    if not (Used.mem p !used) && place i j p then raise Exit)
      with Exit ->
	());
  display_poss ();
  print_score ()
****)

let search_for i j =
  printf "todo...@."
(******
  printf "search...@.";
  let r = 
    if j < n-1 && sol.(i).(j+1) != dummy_piece then sol.(i).(j+1).left else -1 
  in
  let l = 
    if j > 0 && sol.(i).(j-1) != dummy_piece then sol.(i).(j-1).right else -1 
  in
  let u = 
    if i < n-1 && sol.(i+1).(j) != dummy_piece then sol.(i+1).(j).down else -1 
  in
  let d = 
    if i > 0 && sol.(i-1).(j) != dummy_piece then sol.(i-1).(j).up else -1 
  in
  let matches p =
    let one_match pr pu pl pd = 
      (r = -1 || pr = r) && (l = -1 || pl = l) && 
      (u = -1 || pu = u) && (d = -1 || pd = d)
    in
    one_match p.right p.up p.left p.down ||
    one_match p.down p.right p.up p.left ||
    one_match p.left p.down p.right p.up ||
    one_match p.up p.left p.down p.right
  in
  iter_grid
    (fun i' j' ->
      if sol.(i').(j') != dummy_piece && matches sol.(i').(j') then begin
	printf "match at %d,%d@." i' j';
	color_cell i' j' red
      end
    )
*****)

(* interaction loop *)

let ij st = 
  let x = st.mouse_x in
  let y = st.mouse_y in
  if x >= step && x < step+n*step && y >= step && y < step+n*step then
    (y - step) / step, (x - step) / step
  else
    -1,-1

let () =
  display ();
  display_poss ();
  while true do
    let st = wait_next_event [Button_down; Key_pressed; Mouse_motion] in
    if st.keypressed then begin match st.key with
      | 'q' -> exit 0
      | 'c' -> clear_todo ()
      | 'b' -> fill_todo_bt ()
      | 'g' -> greedy ()
      | 'p' -> display_poss ()
      | 'u' -> pop ()
      | 's' -> save !board "saved.sol"; printf "saved in saved.sol@."
      | 'o' -> let i,j = ij st in if i <> -1 then search_for i j
      | 'x' -> let i,j = ij st in if i <> -1 then
	  begin
	    todo := Todo.remove i j !todo;
	    display_poss();
	    display_todo()
	  end
(****
      | 'r' -> 
	  let i,j = ij st in if i <> -1 then begin
	    remove i j;
	    display_poss ();
	    print_score ()
	  end
*****)
      | _ -> ()
    end;
    if st.button then begin
      let i,j = ij st in
      if i >= 0 && piece_at !board i j == dummy_piece then begin
	todo := Todo.add i j !todo;
	color_cell i j yellow
      end
    end
  done
