
open Format
open Brute7
open Network

let current_tree = ref (init_tree ())

let save_current_tree () =
  let t = !current_tree in
  let n = size t in 
  let f = "server_dump_" ^(string_of_int n) in
  dump_tree (Filename.concat sol_dir f) t;
  printf "saved dump file %s@." f

let print_status fmt st =
  match st with
    | None -> fprintf fmt "idle"
    | Some(it,tile,rot,sc) -> 
	fprintf fmt "iteration=%d, tile=%d, rot=%d, score=%d" it tile rot sc

  
let client_table = Hashtbl.create 7

let time = ref (Unix.time ())

module Game = struct

  let new_client c = 
    printf "Got new client, id=%d@." c.Client.id;
    Hashtbl.add client_table c.Client.id (c,ref None); true

  type message = computing_client_answer

  let receive c m = 
    match m with
      | Report(tile,rot,score,size) ->
	  printf "Client %d reports: iteration %d, on tile=%d, rot=%d, score=%d@." c.Client.id size tile rot score;	  
	  let (_,st) = Hashtbl.find client_table c.Client.id in
	  st := Some(size,tile,rot,score)
      | Tree t -> 
	  printf "Received tree from client %d@." c.Client.id;	  
	  current_tree := merge !current_tree (build_tree false t);
	  let ti = Unix.time () in
	  if ti -. !time > 60.0 (* 1 minutes *) then
	    begin
	      save_current_tree();
	      time := ti;
	    end
      | Answer m -> 
	  printf "Received message from client %d: %s@." c.Client.id m
      | Stopped ->
	  printf "Client %d has stopped computing, now idle@." c.Client.id;
	  let (_,st) = Hashtbl.find client_table c.Client.id in
	  st := None
	    

  let each_tick () = ()

  let send c = () 

  let disconnection c = 
    printf "Client disconnected, id=%d@." c.Client.id;
    Hashtbl.remove client_table c.Client.id
    

end

module Server = Network.Make(Game)

(* main launch *)

let () =
  begin
    printf "Starting server@.";
    let server_pid = 
      Thread.create
	(fun () -> Server.run ~port:2007 ~ack:false ~timeout:(1.0))
	()
    in
    while true do
      printf "-------------------------------@.";
      printf "1) show clients' status@.";
      printf "2) show current tree@.";
      printf "3) send computation to a client@.";
      printf "4) stop computation of a client@.";
      printf "5) resort tree@.";
      printf "6) save tree@.";
      printf "10) quit@.";
      let s = read_line() in
      match s with
	| "1" ->
	    printf "-------------------------------@.";
	    Hashtbl.iter 
	      (fun id (c,st) ->
		 printf "client %2d: status %a@." id print_status !st)
	      client_table
	| "2" ->
	    display_stats ~prof:0 !current_tree
	| "3" ->
	    begin
	      try
		printf "client id:@.";
		let id = int_of_string(read_line()) in
		printf "tile:@.";
		let tile = int_of_string (read_line()) in
		printf "rot:@.";
		let rot = int_of_string (read_line()) in
		let (c,st) = Hashtbl.find client_table id in
		printf "send computation for (%d,%d) to client %d@." 
		  tile rot id;
		Network.send c.Client.fd (DoCompute(!current_tree,tile,rot))
	      with _ -> 
		printf "failed to send computation@."
	    end
	| "4" ->
	    begin
	      try
		printf "client id:@.";
		let id = int_of_string(read_line()) in
		let (c,st) = Hashtbl.find client_table id in
		printf "stop computation of client %d@." id;
		Network.send c.Client.fd StopComputation
	      with _ -> 
		printf "failed to stop computation@."
	    end
	| "5" ->
	    current_tree := build_tree true (tree_to_ast !current_tree);
	    display_stats ~prof:0 !current_tree
	| "6" -> save_current_tree()
	| "10" ->
	    Thread.kill server_pid;
	    exit 0
	| _ -> 
	    printf "unknown choice: %s@." s
    done
      
  end
      
