
(* graphical editor *)

open Format
open Graphics
open Puzzle
open Board

let file = ref None

let () = 
  Arg.parse 
    ["-p", Arg.String (fun _ -> ()), ""]
    (fun f -> file := Some f) "editor [-p file.data] [file.sol]"
let file = !file

let step = 55
let half = step / 2
let scale = half-1

let () = 
  open_graph (sprintf " %dx%d" ((2*n+3) * step) ((n+2) * step));
  set_color black;
  for i = 0 to n do
    moveto step (step+i*step); lineto (step+n*step) (step+i*step);
    moveto (step+i*step) step; lineto (step+i*step) (step+n*step)
  done

let draw_tile i j t = 
  let x = step + j * step + half in
  let y = step + i * step + half in
  Draw.piece x y scale t

let color_cell i j c =
  let x = step + j * step in
  let y = step + i * step in
  set_color c;
  fill_rect (x+1) (y+1) (step-2) (step-2)

let gris = rgb 150 150 150

let init_board = match file with
  | None ->
      Board.initial
  | Some file ->
      let init_sol = Puzzle.load file in
      Board.create ~complete:false init_sol

let board = ref init_board

let iter_grid f =
  for i = 0 to n-1 do for j = 0 to n-1 do f i j done done

let display () =
  iter_grid 
    (fun i j -> 
      let p = piece_at !board i j in
      if p != dummy_piece then draw_tile i j p else color_cell i j white)

let best = ref 0

let dump score =
  let file = Filename.concat "edit-manual" (string_of_int score ^ ".sol") in
  if is_eternity_ii && not (Sys.file_exists file) then begin
    save !board file;
    printf "saved.@."
  end

let print_score () =
  let score = score !board in
  printf "score = %d@." score;
  if score > !best then begin
    best := score;
    if score > 200 then dump score
  end

let nbposs = Array.create_matrix n n 0

let display_poss () =
  let pot = ref 0 in
  iter_empty
    (fun i j ->
      color_cell i j white;
      let c = Pset.cardinal (poss !board i j) in
      if c = 0 then color_cell i j gris;
      set_color blue;
      moveto (step + j*step + 5) (step + i*step + 5);
      draw_string (string_of_int c);
      nbposs.(i).(j) <- c)
    !board;
  (* maximal possible score *)
  iter_empty
    (fun i j ->
      if nbposs.(i).(j) > 0 then begin
	let sol = piece_at !board in
	if i > 0 then 
 	  if sol (i-1) (j) == dummy_piece then 
	    (if nbposs.(i-1).(j) > 0 then incr pot)
	  else pot := !pot+2;
	if i < n-1 then
 	  if sol (i+1) (j) == dummy_piece then 
	    (if nbposs.(i+1).(j) > 0 then incr pot)
	  else pot := !pot+2;
	if j > 0 then 
 	  if sol (i) (j-1) == dummy_piece then 
	    (if nbposs.(i).(j-1) > 0 then incr pot)
	  else pot := !pot+2;
	if j < n-1 then
 	  if sol (i) (j+1) == dummy_piece then 
	    (if nbposs.(i).(j+1) > 0 then incr pot)
	  else pot := !pot+2
      end)
    !board;
  assert (!pot mod 2 = 0);
  printf "potential score = %d@." (score !board + !pot/2)

(* board stack *)

let stack = ref []

let push () = 
  stack := !board :: !stack

let pop () = match !stack with
  | [] -> 
      ()
  | b :: st -> 
      stack := st;
      board := b;
      display ();
      display_poss ()

(* interaction loop *)

let ij st = 
  let x = st.mouse_x in
  let y = st.mouse_y in
  if y >= step && y < step+n*step then
    (y - step) / step, (x - step) / step
  else
    -1,-1

let current_sel = ref None
let current_pos = Array.create_matrix n (4*n) dummy_piece
let last_col = ref 0
let last_row = ref 0

let selection i j =
  begin match !current_sel with None -> () | Some (i,j) -> display_poss () end;
  current_sel := Some (i,j);
  color_cell i j yellow;
  set_color white;
  fill_rect ((n+2)*step) step (n*step) (n*step);
  let ps = poss !board i j in
  last_col := 0;
  last_row := 0;
  Pset.iter
    (fun p ->
       current_pos.(!last_row).(!last_col) <- p;
       draw_tile !last_row (n + 1 + !last_col) p;
       incr last_row;
       if !last_row = n then begin incr last_col; last_row := 0 end)
    ps

let pickup r c =
  let p = current_pos.(r).(c) in
  try
    begin match !current_sel with
    | Some (i,j) ->
	push ();
	board := Board.place ~complete:true !board i j p;
	color_cell i j white; 
	draw_tile i j p;
	last_col := 0;
	last_row := 0;
	current_sel := None;
	display_poss ();
	print_score ()
    | None -> 
	assert false
    end
  with Incomplete ->
    color_cell r (c+n+1) red

let () =
  display ();
  display_poss ();
  while true do
    let st = wait_next_event [Button_down; Key_pressed; Mouse_motion] in
    if st.keypressed then begin match st.key with
      | 'q' -> exit 0
      | 'p' -> display_poss ()
      | 'u' -> pop ()
      | 's' -> save !board "saved.sol"; printf "saved in saved.sol@."
      | _ -> ()
    end;
    if st.button then begin
      let i,j = ij st in
      if i >= 0 && i <= n-1 && piece_at !board i j == dummy_piece then 
	selection i j
      else if j >= n+1 then
	let c = j-n-1 in
	if c < !last_col || (c = !last_col && i < !last_row) then pickup i c
    end
  done
