
open Printf
open Unix

let debug = ref true

let dmsg s = 
  if !debug then begin Printf.eprintf "%s" s; flush Pervasives.stderr end

module Client = struct

  type t = { fd : Unix.file_descr; id : int; mutable ready : bool }

  let equal c1 c2 = c1.fd == c2.fd
  let hash c = Hashtbl.hash c.fd
  let compare c1 c2 = Pervasives.compare c1.fd c2.fd

end

module type Game = sig
  val new_client : Client.t -> bool
  type message 
  val receive : Client.t -> message -> unit
  val each_tick : unit -> unit
  val send : Client.t -> unit
  val disconnection : Client.t -> unit
end

let hostname s = 
  try
    let h = gethostbyname s in h.h_addr_list.(0)
  with _ -> try
    (* may be the address was passed as an IP number *)
    inet_addr_of_string s
  with _ -> try
    let h = gethostbyname "localhost" in h.h_addr_list.(0)
  with _ ->
    failwith "Netgame.hostname"

let hs = Marshal.header_size
let header = String.create hs

(*
let rec read_messages handler fd = 
  let len = Marshal.total_size header 0 in 
  let buf = String.create (len+hs) in 
  String.blit header 0 buf 0 hs; 
  let rcv = recv fd buf hs len [] in 
  if !debug then 
    (eprintf "read_messages: rcv=%d, len=%d\n" rcv len; flush Pervasives.stderr);
  let obj = Marshal.from_string buf 0 in
  if rcv < len then 
    handler obj
  else begin
    String.blit buf len header 0 hs; 
    handler obj;
    read_messages handler fd
  end
*)

let read_message handler fd = 
  let len = Marshal.total_size header 0 in 
  let buf = String.create len in 
  String.blit header 0 buf 0 hs; 
  let rec fill offset n =
    if !debug then 
      (eprintf "read_message: offset=%d, n=%d\n" offset n; flush Pervasives.stderr);
    let rcv = recv fd buf offset n [] in
    if !debug then 
      (eprintf "read_message: rcv=%d\n" rcv ; flush Pervasives.stderr);
    if rcv < n then fill (offset+rcv) (n-rcv) else
      if rcv > n then failwith "overflow" else
	()
  in 
  fill hs (len-hs);
  let obj = Marshal.from_string buf 0 in
  handler obj


module Make(G : Game) = struct

  module H = Hashtbl

  (* Clients are stored in the following hash table. It maps each client
     to a boolean reference indicating if it is ready. *)
  let clients = (H.create 17 : (file_descr, Client.t) H.t)

  let client = H.find clients

  let sock = socket PF_INET SOCK_STREAM 0

  let sockets_list = ref ([sock] : file_descr list)

  open Client

  let create_client = 
    let r = ref 0 in
    fun fd -> incr r; { fd = fd; id = !r; ready = true }

  let new_client () =
    let fd,_ = accept sock in
    dmsg "new connection...";
    let c = create_client fd in
    if G.new_client c then begin
      dmsg ("accepted as client " ^ string_of_int c.id ^ "\n");
      H.add clients fd c;
      sockets_list := fd :: !sockets_list
    end else
      dmsg "rejected\n"

  let disconnection c =
    dmsg ("client " ^ string_of_int c.id ^ " disconnected\n");
    G.disconnection c;
    H.remove clients c.fd;
    sockets_list := List.filter ((!=) c.fd) !sockets_list

  let receive c =
    try
      if (recv c.fd header 0 hs []) < hs then disconnection c
      else read_message (fun m -> c.ready <- true; G.receive c m) c.fd
    with Unix.Unix_error (errno,_,_) as e ->
      (* TODO: call disconnection whatever the error is? *)
      if errno = ECONNRESET then 
	disconnection c
      else if !debug then begin
	eprintf "Netgame.Make(G).receive failed (%s)\n" (Printexc.to_string e);
	flush Pervasives.stderr
      end

  let run ~port ~ack ~timeout =
    begin try
      let machine = gethostname() in
      let sockaddr = ADDR_INET (hostname machine, port) in
      setsockopt sock SO_REUSEADDR true;
      bind sock sockaddr;
      listen sock 3
    with Unix.Unix_error _ -> 
      eprintf "run: couldn't run the server\n"; exit 1
    end;
    let handle s = if s == sock then new_client () else receive (client s) in
    while true do
      if ack then Hashtbl.iter (fun _ c -> c.ready <- false) clients;
      let ready,_,_ = select !sockets_list [] [] timeout in
      List.iter handle ready;
      G.each_tick ();
      Hashtbl.iter (fun _ c -> if c.ready then G.send c) clients
    done

end

(* Client helper functions *)

let connection ~machine ~port =
  try
    let sock = socket PF_INET SOCK_STREAM 0 in
    let machine = if machine="" then gethostname() else machine in
    let sockaddr = ADDR_INET(hostname machine, port) in
    connect sock sockaddr;
    sock
  with Unix.Unix_error(errno, msg1, msg2) ->
    eprintf "err=%s, msg1=%s, msg2=%s@." (Unix.error_message errno) msg1 msg2;
    failwith ("connection: couldn't connect to the server " ^ machine ^
		":" ^ string_of_int port)

let send server m =
  let s = Marshal.to_string m [] in
  let len = String.length s in
  if !debug then (eprintf "send: len=%d\n" len; flush Pervasives.stderr);
  let rec do_send offset len =
    if len <= 16384 then
      let res = send server s offset len [] in
      if !debug then (eprintf "send: res=%d\n" res; flush Pervasives.stderr);
    else
      let res = send server s offset 16384 [] in
      if !debug then (eprintf "send: res=%d\n" res; flush Pervasives.stderr);
      do_send (offset+16384) (len-16384)
  in
  do_send 0 len

exception Shutdown

let receive ~timeout handler server =
  try
    let ready,_,_ = Unix.select [server] [] [] timeout in
    if ready <> [] then
      if (recv server header 0 hs []) < hs then raise Shutdown
      else read_message handler server
  with Unix.Unix_error _ as e ->
    if !debug then begin
      Printf.eprintf "Network.receive failed (%s)\n" (Printexc.to_string e);
      flush Pervasives.stderr
    end
