
(* Server side *)

val debug : bool ref 

module Client : sig

  type t = private { fd : Unix.file_descr; id : int; mutable ready : bool }

  val equal : t -> t -> bool
  val hash : t -> int
  val compare : t -> t -> int

end

module type Game = sig

  val new_client : Client.t -> bool
    (** [new_client c] is called on any new client attempting a connection.
	The boolean indicates whether this client is allowed to enter the 
	game. *)

  type message 
    (** the type of messages sent from the clients to the server *)

  val receive : Client.t -> message -> unit
    (** [receive c m] is applied each time a message [m] is received from
	client [c] *)

  val each_tick : unit -> unit
    (** this function is called once each step, after listening
        to clients and before sending them messages;
        it is usually the function updating the game logic;
        it is also the function waiting some appropriate delay to achieve
        a given game frequency *)

  val send : Client.t -> unit
    (** this function is called for any (ready) client, each step *)

  val disconnection : Client.t -> unit
    (** [disconnection c] is called when client [c] disconnects *)

end

module Make(G : Game) : sig

  val run : port:int -> ack:bool -> timeout:float -> unit
    (** launches a server on the given port.
	The server enters an infinite loop where at each round it
        - listens for messages from clients and possibly new clients
	  (with [G.receive] and [G.new_client]; [timeout] is the maximum 
	  delay, in seconds to wait for messages (if negative, waits 
	  indefinitely)  
        - calls [G.each_ticks];
	- sends messages to all clients with [G.send].
	  When [ack] is set, messages are only sent to clients considered
	  ready. A client is considered ready when it has sent some message 
	  to the server in the current round. So when ack is on, it is 
	  important for clients to send messages to the server at each round 
	  (even if they have nothing to say); see example in test.ml *)

end

(* Helper functions for the client side *)

val connection : machine:string -> port:int -> Unix.file_descr

val send : Unix.file_descr -> 'a -> unit

exception Shutdown

val receive : timeout:float -> ('a -> unit) -> Unix.file_descr -> unit
  (** [receive h fd] receives Ocaml values on socket [fd] and applies
      function [h] on all of them (sequentially).  [timeout] is the
      maximum reception delay, in seconds, to wait for messages (if
      negative, waits indefinitely).
      raises [Shutdown] if the socket is closed *)

