
(* reading [n] and the [n*n] pieces from file or "eternity-ii.data" *)

open Format

let f =
  let rec scan = function
    | [] -> "eternity-ii.data"
    | "-p" :: f :: _ ->	f 
    | _ :: args -> scan args
  in
  scan (Array.to_list Sys.argv)

let ch = 
  if Sys.file_exists f then begin
    let ch = open_in f in
    at_exit (fun () -> close_in ch);
    ch
  end else begin
    eprintf "cannot open file %s@." f;
    exit 1
  end

let is_eternity_ii = input_line ch = "eternity-ii"

let n = int_of_string (input_line ch)

let n2 = n*n

let corner_distance i j = min i (n-i) + min j (n-j)
    
type kind = Corner | Edge | Other

type piece = 
  { id : int; kind : kind; 
    right : int; up : int; left : int; down : int;
    rot : int;
    uid : int; (* unique id = (id lsl 2) lor rot *)
  }

(* counter clockwise rotation *)
let rot p =
  let r = (p.rot + 1) mod 4 in
  { p with 
      right = p.down; up = p.right; left = p.up; down = p.left;
      rot = r; uid = (p.id lsl 2) lor r }


let dummy_piece = 
  { id= -1; kind=Other; right=0; up=0; left=0; down=0; rot=0; uid = -1 }

let dropped_piece =
  { id= -2; kind=Other; right=0; up=0; left=0; down=0; rot=0; uid = -2 }

let pieces = Array.create (n*n) dummy_piece (* initial orientation *)
let canonical_pieces = Array.create (n*n) dummy_piece

exception BadPiece

let is_real_piece p = p.id >= 0 (* p != dummy_piece && p != dropped_piece *)

let make_piece i r u l d =
  let gray = ref 0 in
  if r = 0 then incr gray;
  if u = 0 then incr gray;
  if l = 0 then incr gray;
  if d = 0 then incr gray;
  let k,(r,u,l,d,rot) = match !gray with
    | 0 -> 
	Other,(r,u,l,d,0)
    | 1 -> 
	Edge, 
	if u = 0 then r,u,l,d,0
	else if l = 0 then u,l,d,r,3
	else if d = 0 then l,d,r,u,2
	else if r = 0 then d,r,u,l,1
	else assert false
    | 2 -> 
	Corner, 
	if u = 0 && l = 0 then r,u,l,d,0
	else if l = 0 && d = 0 then u,l,d,r,3
	else if d = 0 && r = 0 then l,d,r,u,2
	else if r = 0 && u = 0 then d,r,u,l,1
	else raise BadPiece
    | _ -> 
	raise BadPiece
  in
  { id=i; kind=k; right=r; up=u; left=l; down=d; rot=rot; 
    uid = (i lsl 2) lor rot }

let () =
  let k = ref 0 in
  try
    while true do
      let s = input_line ch in
      Scanf.sscanf s "%d %d %d %d %d" 
	(fun i r u l d -> 
	  incr k;
	  let p = canonical_pieces.(i-1) in 
	  if p.id <> -1 then begin 
	    eprintf "duplicate piece %d@." i; exit 1 
	  end;
	  let t = make_piece i r u l d in
	  canonical_pieces.(i-1) <- t;
	  pieces.(i-1) <- 
	    { id=i; kind=t.kind; right=r; up=u; left=l; down=d; rot=0; 
	      uid = i lsl 2 })
    done
  with 
    | End_of_file -> 
	if !k <> n*n then begin eprintf "bad input@."; exit 1 end
    | Scanf.Scan_failure s -> 
	eprintf "Scan_failure: %s@." s; exit 1
    | BadPiece ->
	eprintf "bad piece on line %d@." (!k + 1)

(* sorting pieces by type *)

let nb_corners = 4
let nb_edges = 4*n-8
let nb_others = (n-2) * (n-2)

let corners_rot0 = Array.create nb_corners dummy_piece
let edges_rot0 = Array.create nb_edges dummy_piece
let others_rot0 = Array.create nb_others dummy_piece

let () =
  let corner = ref 0 in
  let edge = ref 0 in
  let other = ref 0 in
  Array.iter
    (fun p -> match p.kind with
      | Corner -> 
	  corners_rot0.(!corner) <- p; 
	  incr corner; 
	  if !corner > nb_corners then (eprintf "too many corners@."; exit 1)
      | Edge ->
	  edges_rot0.(!edge) <- p; 
	  incr edge; 
	  if !edge > nb_edges then (eprintf "too many edges@."; exit 1)
      | Other ->
	  others_rot0.(!other) <- p; 
	  incr other; 
	  if !other > nb_others then (eprintf "too many others@."; exit 1)
    )
    canonical_pieces;
  assert (!corner = nb_corners && !edge = nb_edges && !other = nb_others)

let corners_rot1 = Array.map rot corners_rot0
let corners_rot2 = Array.map rot corners_rot1
let corners_rot3 = Array.map rot corners_rot2

let edges_rot1 = Array.map rot edges_rot0
let edges_rot2 = Array.map rot edges_rot1
let edges_rot3 = Array.map rot edges_rot2

let others_rot1 = Array.map rot others_rot0
let others_rot2 = Array.map rot others_rot1
let others_rot3 = Array.map rot others_rot2

(* sanity checks *)

let occ = Hashtbl.create 97

let add_occ p = 
  try let n = Hashtbl.find occ p in Hashtbl.replace occ p (n+1)
  with Not_found -> Hashtbl.add occ p 1

let fill = 
  Array.iter 
    (fun p -> add_occ p.right; add_occ p.up; add_occ p.left; add_occ p.down)

let () = fill corners_rot0; fill edges_rot0; fill others_rot0

let () = Hashtbl.iter (fun _ n -> assert (n mod 2 = 0)) occ

(* statistics *)

(*
let () = 
  Hashtbl.iter (fun p n -> printf "pattern %d: %d occurrences@." p n) occ
*)

(*
let () =
  Array.iter
    (fun p -> printf "%d@ " p.down)
    edges_rot0;
  printf "@."
*)

(* save/load *)

let pieces_rot = Array.create 4 [||]

let () = pieces_rot.(0) <- pieces
let () = pieces_rot.(1) <- Array.map rot pieces_rot.(0)
let () = pieces_rot.(2) <- Array.map rot pieces_rot.(1)
let () = pieces_rot.(3) <- Array.map rot pieces_rot.(2)

let get_piece pid r = pieces_rot.(r).(pid-1)

let load file =
  let s = Array.create_matrix n n (-1,0) in
  Solution.load n file s;
  let sol = Array.create_matrix n n dummy_piece in
  for i = 0 to n-1 do 
    for j = 0 to n-1 do
      let (p,r) = s.(i).(j) in
      if p >= 0 then
	sol.(i).(j) <- get_piece p r
    done 
  done;
  sol

let save file sol =
  let s = Array.create_matrix n n (-1,0) in
  for i = 0 to n-1 do 
    for j = 0 to n-1 do
      let p = sol.(i).(j) in
      s.(i).(j) <- (p.id,p.rot)
    done 
  done;
  Solution.save n s file
