
(* reading [n] and the [n*n] tiles from file or "eternity-ii.data" *)

open Format

let f =
  let rec scan = function
    | [] -> "eternity-ii.data"
    | "-p" :: f :: _ ->	f 
    | _ :: args -> scan args
  in
  scan (Array.to_list Sys.argv)

let ch = 
  if Sys.file_exists f then begin
    let ch = open_in f in
    at_exit (fun () -> close_in ch);
    ch
  end else begin
    eprintf "cannot open file %s@." f;
    exit 1
  end

let is_eternity_ii = input_line ch = "eternity-ii"

let n = int_of_string (input_line ch)

let n2 = n*n

type kind = Corner of int | Edge of int | Other
    (* arg of Corner and Edge gives the correct rotation if placed at (0,0) *)

type tile = 
  { id : int; kind : kind; 
    pats : int array (* right, up, left, down *)
  }

let right_pat p r = p.pats.((4-r) mod 4)
let up_pat p r = p.pats.((5-r) mod 4)
let left_pat p r = p.pats.((6-r) mod 4)
let down_pat p r = p.pats.((7-r) mod 4)

let dummy_tile = 
  { id= -1; kind=Other; pats = [||] }

let tiles = Array.create (n*n) dummy_tile 

exception BadTile

let is_real_tile p = p.id >= 0

let make_tile i r u l d =
  let gray = ref 0 in
  if r = 0 then incr gray;
  if u = 0 then incr gray;
  if l = 0 then incr gray;
  if d = 0 then incr gray;
  let k,(p0,p1,p2,p3) = match !gray with
    | 0 -> 
	Other,(r,u,l,d)
    | 1 -> 
	let rot = 
	  if u = 0 then 2
	  else if l = 0 then 1
	  else if d = 0 then 0
	  else if r = 0 then 3
	  else assert false
	in	  
	(Edge rot), (r,u,l,d)
    | 2 -> 
	let rot = 
	  if u = 0 && l = 0 then 1
	  else if l = 0 && d = 0 then 0
	  else if d = 0 && r = 0 then 3
	  else if r = 0 && u = 0 then 2
	  else raise BadTile
	in
	(Corner rot), (r,u,l,d)
    | _ -> 
	raise BadTile
  in
  { id=i; kind=k; pats=[|p0;p1;p2;p3|]; }

let () =
  let k = ref 0 in
  try
    while true do
      let s = input_line ch in
      Scanf.sscanf s "%d %d %d %d %d" 
	(fun i r u l d -> 
	  incr k;
	  let p = tiles.(i-1) in 
	  if p.id <> -1 then begin 
	    eprintf "duplicate tile %d@." i; exit 1 
	  end;
	  let t = make_tile i r u l d in
	  tiles.(i-1) <- t)
    done
  with 
    | End_of_file -> 
	if !k <> n*n then begin eprintf "bad input@."; exit 1 end
    | Scanf.Scan_failure s -> 
	eprintf "Scan_failure: %s@." s; exit 1
    | BadTile ->
	eprintf "bad tile on line %d@." (!k + 1)

(* sorting tiles by type *)

let nb_corners = 4
let nb_edges = 4*n-8
let nb_others = (n-2) * (n-2)

let corners = Array.create nb_corners dummy_tile
let edges = Array.create nb_edges dummy_tile
let others = Array.create nb_others dummy_tile

let () =
  let corner = ref 0 in
  let edge = ref 0 in
  let other = ref 0 in
  Array.iter
    (fun p -> match p.kind with
      | Corner _ -> 
	  corners.(!corner) <- p; 
	  incr corner; 
	  if !corner > nb_corners then (eprintf "too many corners@."; exit 1)
      | Edge _ ->
	  edges.(!edge) <- p; 
	  incr edge; 
	  if !edge > nb_edges then (eprintf "too many edges@."; exit 1)
      | Other ->
	  others.(!other) <- p; 
	  incr other; 
	  if !other > nb_others then (eprintf "too many others@."; exit 1)
    )
    tiles;
  assert (!corner = nb_corners && !edge = nb_edges && !other = nb_others)

(* sanity checks *)

let occ = Hashtbl.create 97

let add_occ p = 
  try let n = Hashtbl.find occ p in Hashtbl.replace occ p (n+1)
  with Not_found -> Hashtbl.add occ p 1

let fill = 
  Array.iter 
    (fun p -> add_occ p.pats.(0); add_occ p.pats.(1); 
       add_occ p.pats.(2); add_occ p.pats.(3))

let () = fill corners; fill edges; fill others

let () = Hashtbl.iter (fun _ n -> assert (n mod 2 = 0)) occ

(* statistics *)

(*
let () = 
  Hashtbl.iter (fun p n -> printf "pattern %d: %d occurrences@." p n) occ
*)

(*
let () =
  Array.iter
    (fun p -> printf "%d@ " p.down)
    edges_rot0;
  printf "@."
*)

