
open Format
open Graphics
open Puzzle0
open Board0

let graphics = ref false
let load = ref None
let sol_dir = ref "recuit"
let spec = 
  ["-p", Arg.String (fun _ -> ()), "<file>  set puzzle file";
   "-graphics", Arg.Set graphics, "with graphical display";
   "-dir", Arg.Set_string sol_dir, "<dir>  directory for solutions";]
let usage = "recuit.opt [-p puzzle] [-graphics] file"
let () = Arg.parse spec (fun f -> load := Some f) usage
  
let graphics = !graphics
let load = match !load with
  | None -> Arg.usage spec usage; exit 1
  | Some f -> f

let sol_dir = !sol_dir


let step = 56
let half = step / 2
let scale = ref (half+1)

let () =
  if graphics then 
    Draw.open_board ~title:("Recuit for file " ^ load) ~lines:false step


let draw_tile i j t r =
  let x = step + j * step + half in
  let y = step + i * step + half in
  Draw.right x y !scale (Puzzle0.right_pat t r);
  Draw.up x y !scale (Puzzle0.up_pat t r);
  Draw.left x y !scale (Puzzle0.left_pat t r);
  Draw.down x y !scale (Puzzle0.down_pat t r)


let clear_tile i j =
  let x = step + j * step in
  let y = step + i * step in
  set_color white;
  fill_rect (x+1) (y+1) (step-2) (step-2)

let border_tile i j =
  let x = step + j * step in
  let y = step + i * step in
  set_color white;
  draw_rect (x+1) (y+1) (step-2) (step-2)

let iter_grid f =
  for i = 0 to n-1 do for j = 0 to n-1 do f i j done done

let draw_board b =
  if graphics then begin
    iter_grid 
      (fun i j -> 
	 let p,r = get b i j in
	 if p != dummy_tile then draw_tile i j p r else clear_tile i j)
  end
  
let init_board = Board0.load load 

let best = ref 0

let display b = 
  draw_board b;
  printf "score = %d@." b.score;
  if b.score > !best then
    begin
      best := b.score;
      let f = (Filename.basename load) ^ string_of_int b.score
      in
      let file = Filename.concat sol_dir f in
      if is_eternity_ii && not (Sys.file_exists file) then begin
	Board0.save b file;
	printf "saved in %s.@." file;
	ignore(read_key())
      end;
    end


(* 

   fills the corners of board b, with tiles which increase the score
   for at least [min]

   If there are no more corners to fill, just display the board and
   wait a key to be pressed.

*)
let rec fill_corners min b =
  try
    List.iter
      (fun (i,j,rot) ->
	 let p,_ = get b i j in
	 if not (is_real_tile p) then 
	   begin
	     let found = ref false in
	     TileSet.iter
	       (fun p ->
		  let r = match p.kind with
		    | Corner r -> (r+rot) mod 4
		    | _ -> assert false
		  in 
		  let b1 = put b i j p r in
		  if b1.score >= b.score+min then 
		    begin
		      found := true;
		      fill_corners min b1;
		    end;
	       )
	       b.remaining_corners;
	     if !found then raise Exit;
	   end)
      [(0,0,0) ; (0,n-1,1) ; (n-1,n-1,2) ; (n-1,0,3) ];
    if b.score > !best then 
      begin
	display b; 
	ignore(read_key())
	  (* Format.printf "full board@."; *)
	  (* ignore(read_key()) *)
      end;
    raise Exit
  with Exit -> 
    if min > 1 then
      begin
	display b; 
	Format.printf "now try filling with min=%d@." (min-1);
	ignore(read_key()); 
	Format.printf "go...@.";
	fill_remaining (min-1) 1 1 b
      end

and fill_border min i0 b =
  try
    let x = ref i0 in
    while !x < n-1 do
      List.iter
	(fun (i,j,rot) ->
	   let p,_ = get b i j in
	   if not (is_real_tile p) then 
	     begin
	       let found = ref false in
	       TileSet.iter
		 (fun p ->
		    let r = match p.kind with
		      | Edge r -> (r+rot) mod 4
		      | _ -> assert false
		    in 
		    let b1 = put b i j p r in
		    if b1.score >= b.score+min then 
		      begin
			found := true;
			fill_border min !x b1;
		      end;
		 )
		 b.remaining_edges;
	       if !found then raise Exit
	     end)
	[(0,!x,0) ; (!x,n-1,1) ; (n-1,!x,2) ; (!x,0,3)];
      incr x;
    done;
    raise Exit
  with Exit -> 
    (* inside and border is full, now fills the corners *)
    fill_corners min b

(* try filling remaining empty slots, inside first, starting from
   (i0,j0) (thus board is assumed fully filled for (i,j) such that
   i<i0 or i=i0 and j < j0) *)
and fill_remaining min i0 j0 b =
  try
    let i = ref i0 and j = ref j0 in
    while !i < n-1 do
      while !j < n-1 do
	let p,_ = get b !i !j in
	if not (is_real_tile p) then 
	  begin
	    let found = ref false in
	    TileSet.iter
	      (fun p ->
		 for r=0 to 3 do
		   let b1 = put b !i !j p r in
		   if b1.score >= b.score+min then 
		     begin
		       found := true;
		       fill_remaining min !i (!j+1) b1;
		     end;
		 done)
	      b.remaining_others;
	    if !found then raise Exit
	  end;
	incr j;
      done;
      incr i;
      j := 1
    done;
    (* no more empty slot to do *)
    raise Exit
  with Exit -> 
    (* inside is already full, now fills the border from start *)
    fill_border min 0 b

let rec loop b =
  let pos = ref [] in
  (* inside *)
  for i = 1 to n-2 do
    for j = 1 to n-2 do
      let p,_ = get b i j in
      if not (is_real_tile p) then 
	TileSet.iter
	  (fun p ->
	     for r=0 to 3 do
	       let s = score_delta b i j p r in
	       if s > 0 then pos := (s,i,j,p,r) :: !pos
	     done)
	  b.remaining_others
    done
  done;
  (* edges *)
  for i = 1 to n-2 do
    List.iter
      (fun (i,j,rot) -> 
	 let p,_ = get b i j in
	 if not (is_real_tile p) then 
	   TileSet.iter
	     (fun p ->
		let r = match p.kind with
		  | Edge r -> (r+rot) mod 4
		  | _ -> assert false
		in 
		let s = score_delta b i j p r in
		if s > 0 then pos := (s,i,j,p,r) :: !pos)
	     b.remaining_edges)
      [(0,i,0) ; (i,n-1,1) ; (n-1,i,2) ; (i,0,3)];
  done;
  (* corners *)
  List.iter
    (fun (i,j,rot) -> 
       let p,_ = get b i j in
       if not (is_real_tile p) then 
	 TileSet.iter
	   (fun p ->
	      let r = match p.kind with
		| Corner r -> (r+rot) mod 4
		| _ -> assert false
	      in 
	      let s = score_delta b i j p r in
	      if s > 0 then pos := (s,i,j,p,r) :: !pos)
	   b.remaining_corners)
    [(0,0,0) ; (0,n-1,1) ; (n-1,n-1,2) ; (n-1,0,3) ];
  match !pos with
    | [] -> if b.score > ! best then display b	
    | l ->
	let l = 
	  List.sort
	    (fun (s1,_,_,_,_) (s2,_,_,_,_) -> Pervasives.compare s2 s1) 
	    l
	in
	List.iter
	  (fun (_,i,j,p,r) ->
	     let b = put b i j p r in loop b)
	  l


let () = 
  printf "press a key to start@.";
  ignore(read_key());
  loop init_board;
  Format.printf "done.@."
